import asyncio
import param
import panel as pn
import json
from copy import deepcopy
from kedro_graphql.client import KedroGraphqlClient
from kedro_graphql.models import Pipeline, Tag, ParameterType, ParameterInput, DataSetInput, TagInput, PipelineInput, PipelineInputStatus

pn.extension('modal')
pn.extension(notifications=True)

SECTION_ITEM_STYLES = {
    'margin-top': 'auto',
    'margin-bottom': 'auto',
    'justify-content': 'left',
    'background-color': '#f0f0f0',
    'padding': '10px',
    'margin': '10px',
    'border-radius': '5px',
    'word-break': 'break-all',
}

class PipelineCloning(pn.viewable.Viewer):
    """
    A Panel component that clones a Kedro pipeline.

    Attributes:
        pipeline (Pipeline): The Kedro pipeline to retry.
        client (KedroGraphqlClient): The client used to interact with the Kedro GraphQL API.
    """

    pipeline = param.ClassSelector(class_=Pipeline)
    client = param.ClassSelector(class_=KedroGraphqlClient)

    def __init__(self, **params):
        super().__init__(**params)

        pn.state.notifications.position = 'top-center'
        
        if not self.pipeline or not self.client:
            self._content = pn.Column(
                pn.pane.Markdown("**Missing required parameters: `pipeline`, or `client`.**"),
                sizing_mode="stretch_width"
            )
        else:
            self.state = deepcopy(self.pipeline)
            
            # Show loading spinner until the dashboard is built
            self._content = pn.Column(
                pn.indicators.LoadingSpinner(value=True, width=50, height=50),
                pn.pane.Markdown("Fetching Data..."),
                sizing_mode='stretch_width'
            )
            
            # Ensures build after panel server is fully loaded to avoid race conditions
            pn.state.onload(lambda: asyncio.create_task(asyncio.to_thread(self.build_component)))
    
    def build_component(self):
        """Builds the pipeline cloning component with datasets, parameters, and tags sections."""
        self.datasets_section = self._create_datasets_section()
        self.parameters_section = self._create_parameters_section()
        self.tags_section = self._create_tags_section()

        stage_button = pn.widgets.Button(name="Clone & Stage", button_type="primary", width=200, height=50)
        run_button = pn.widgets.Button(name="Clone & Run", button_type="success", width=200, height=50)

        run_button.on_click(lambda event: asyncio.create_task(self.clone_pipeline(type="ready")))
        stage_button.on_click(lambda event: asyncio.create_task(self.clone_pipeline(type="staged")))

        self.view = pn.Column(
            self.datasets_section,
            self.parameters_section,
            self.tags_section,
            pn.Row(
                stage_button,
                run_button,
                styles={'margin-top': '25px'}
            ),
            margin=(10, 10),
            scroll=True
        )
        
        self._content[:] = pn.Column(
            pn.Card(
                self.view,
                title=f"Clone Pipeline: {self.pipeline.name}",
                sizing_mode="stretch_width"
            ),
            sizing_mode="stretch_width"
        )

    def _create_datasets_section(self):
        """Creates a section for editing datasets in the pipeline."""

        widgets = []
        if self.state.data_catalog:
            for dataset in self.state.data_catalog:
                try:
                    config = json.loads(dataset.config)
                    name = dataset.name
                    filepath = config.get("filepath", "")

                    edit_button = pn.widgets.Button(
                        name="Edit",
                        button_type="primary",
                        width=100,
                        styles={'margin-top': 'auto', 'margin-bottom': 'auto'}
                    )
                    modal = self._dataset_edit_modal(dataset, edit_button)

                    dataset_block = pn.Row(
                        edit_button,
                        pn.Column(
                            pn.pane.HTML(f"<b>Name</b>: {name}", styles={'font-size': '16px'}),
                            pn.pane.HTML(f"<b>Filepath</b>: {filepath}", styles={'font-size': '16px'}),
                            styles={'margin-top': 'auto', 'margin-bottom': 'auto'}
                        ),
                        modal,
                        styles=SECTION_ITEM_STYLES
                    )
                    widgets.append(dataset_block)
                except json.JSONDecodeError as e:
                    pn.state.notifications.error(
                        f"Failed to decode JSON for dataset {dataset.name}: {str(e)}",
                        duration=10000
                    )
                    continue

        return pn.Column(
            pn.pane.Markdown("## Datasets"),
            pn.FlexBox(*widgets) if widgets else pn.pane.Markdown("No datasets found."),
        )

    def _dataset_edit_modal(self, dataset, trigger):
        """Return modal for editing dataset configurations"""

        try:
            config = json.loads(dataset.config)
        except json.JSONDecodeError:
            self._show_error(f"Invalid configuration for dataset {dataset.name}")
            return pn.Modal("Error", "Invalid dataset configuration")

        original_path = config.get("filepath", "")
        filepath_input = pn.widgets.TextInput(name="Filepath", value=original_path)

        tag_inputs = []
        modal_items = [filepath_input, "### Dataset Tags"]
        if dataset.tags:
            for tag in dataset.tags:
                delete_button = pn.widgets.Button(name="Delete", button_type="danger", width=100)
                modal = self._delete_dataset_tag_modal(dataset, tag, delete_button)
                tag_key_input = pn.widgets.TextInput(name="Tag Key", value=tag.key)
                tag_value_input = pn.widgets.TextInput(name="Tag Value", value=tag.value)
                tag_inputs.append(pn.Row(delete_button, pn.Column(tag_key_input, tag_value_input), modal, styles={
                    'border': '1px solid #cccccc', 'padding': '10px', 'margin': '10px', 'border-radius': '5px'}))

            for row in tag_inputs:
                modal_items.extend([row])

        save_button = pn.widgets.Button(name="Save", button_type="success")
        cancel_button = pn.widgets.Button(name="Cancel")

        add_button = pn.widgets.Button(name="Add Tag", button_type="success", width=100)
        add_tag_modal = self._add_dataset_tag_modal(dataset, add_button)
        modal_items.append(pn.Row(add_button, add_tag_modal))
        modal_items.append(pn.Row(cancel_button, save_button, styles={'margin-top': '25px'}))

        modal = pn.Modal("### Edit Dataset", *modal_items)

        def open_modal(event):
            modal.show()

        def save(event):
            new_path = filepath_input.value.strip()
            if not new_path or new_path == original_path:
                pn.state.notifications.error("Filepath must be changed and not empty.", duration=10000)
                return

            config["filepath"] = new_path
            dataset.config = json.dumps(config)

            modal.hide()
            self.view[0] = self._create_datasets_section()

        def cancel(event):
            filepath_input.value = original_path
            modal.hide()

        trigger.on_click(open_modal)
        save_button.on_click(save)
        cancel_button.on_click(cancel)

        return modal

    def _delete_dataset_tag_modal(self, dataset, tag, trigger):
        """Return modal for deleting a dataset tag"""

        confirm_button = pn.widgets.Button(name="Confirm", button_type="danger")
        cancel_button = pn.widgets.Button(name="Cancel")

        modal = pn.Modal("### Delete Tag", "Are you sure you want to delete this tag?",
                         pn.Row(cancel_button, confirm_button))

        def open_modal(event):
            modal.show()

        def confirm(event):
            for i, t in enumerate(dataset.tags):
                if t.key == tag.key and t.value == tag.value:
                    del dataset.tags[i]
                    self.view[0] = self._create_datasets_section()
                    pn.state.notifications.success("Tag deleted successfully.", duration=10000)
                    break

            modal.hide()

        def cancel(event):
            modal.hide()

        trigger.on_click(open_modal)
        confirm_button.on_click(confirm)
        cancel_button.on_click(cancel)

        return modal

    def _add_dataset_tag_modal(self, dataset, trigger):
        """Return modal for adding a new tag to a dataset"""

        key_input = pn.widgets.TextInput(name="Key", placeholder="Enter tag key")
        value_input = pn.widgets.TextInput(name="Value", placeholder="Enter tag value")

        save_button = pn.widgets.Button(name="Save", button_type="success")
        cancel_button = pn.widgets.Button(name="Cancel")
        modal = pn.Modal("### Add Tag", key_input, value_input, pn.Row(cancel_button, save_button))

        def open_modal(event):
            modal.show()

        def save(event):
            new_key = key_input.value.strip()
            new_value = value_input.value.strip()
            if new_key and new_value:
                if dataset.tags:
                    dataset.tags.append(Tag(key=new_key, value=new_value))
                else:
                    dataset.tags = [Tag(key=new_key, value=new_value)]
            else:
                pn.state.notifications.error("Key or value cannot be empty.", duration=10000)

            modal.hide()
            self.view[0] = self._create_datasets_section()

        def cancel(event):
            key_input.value = ""
            value_input.value = ""
            modal.hide()

        trigger.on_click(open_modal)
        save_button.on_click(save)
        cancel_button.on_click(cancel)

        return modal

    def _create_parameters_section(self):
        """Creates a section for editing/deleting parameters in the pipeline."""

        widgets = []
        if self.state.parameters:
            for parameter in self.state.parameters:
                edit_button = pn.widgets.Button(name="Edit", button_type="primary", width=100, styles={
                    'margin-top': 'auto', 'margin-bottom': 'auto'})
                modal = self._param_edit_modal(parameter, edit_button)
                widgets.append(pn.Row(
                    edit_button,
                    pn.Column(pn.pane.HTML(f"<b>Name</b>: {parameter.name}", styles={'font-size': '16px'}),
                              pn.pane.HTML(f"<b>Value</b>: {parameter.value}", styles={'font-size': '16px'}),
                              pn.pane.HTML(f"<b>Type</b>: {parameter.type.value}", styles={'font-size': '16px'}), styles={
                        'margin-top': 'auto', 'margin-bottom': 'auto'}),
                    modal, styles=SECTION_ITEM_STYLES
                ))

        return pn.Column(
            pn.pane.Markdown("## Parameters"),
            pn.FlexBox(
                *widgets) if widgets else pn.pane.Markdown("No parameters found."),
        )

    def _param_edit_modal(self, parameter, trigger):
        """Return modal for editing a parameter's value and type"""

        value_input = pn.widgets.TextInput(name="Value", value=parameter.value)
        type_input = pn.widgets.Select(
            name="Type", options=["string", "float", "boolean", "integer"],
            value=parameter.type.value)

        save_button = pn.widgets.Button(name="Save", button_type="success")
        cancel_button = pn.widgets.Button(name="Cancel")
        modal = pn.Modal(f"### Edit Parameter '{parameter.name}'", value_input, type_input, pn.Row(
            cancel_button, save_button, styles={'margin-top': '25px'}))

        def open_modal(event):
            modal.show()

        def save(event):
            new_value = value_input.value.strip()
            new_type = type_input.value
            if new_value:
                parameter.value = new_value
                if new_type == "float":
                    parameter.type = ParameterType.FLOAT
                elif new_type == "boolean":
                    parameter.type = ParameterType.BOOLEAN
                elif new_type == "integer":
                    parameter.type = ParameterType.INTEGER
                else:
                    parameter.type = ParameterType.STRING
                modal.hide()
                self.view[1] = self._create_parameters_section()

            else:
                pn.state.notifications.error("Name, value or type cannot be empty.", duration=10000)

        def cancel(event):
            value_input.value = parameter.value
            type_input.value = parameter.type
            modal.hide()

        trigger.on_click(open_modal)
        save_button.on_click(save)
        cancel_button.on_click(cancel)

        return modal

    def _create_tags_section(self):
        """Creates a section for editing/deleting tags in the pipeline."""

        widgets = []
        if self.state.tags:
            for tag in self.state.tags:
                edit_button = pn.widgets.Button(name="Edit", button_type="primary", width=100)
                edit_tag_modal = self._tag_edit_modal(tag, edit_button)
                delete_button = pn.widgets.Button(name="Delete", button_type="danger", width=100)
                delete_tag_modal = self._pipeline_tag_delete_modal(tag, delete_button)
                widgets.append(
                    pn.Row(
                        pn.Column(edit_button, delete_button, styles={'margin-top': 'auto', 'margin-bottom': 'auto'}),
                        pn.Column(
                            pn.pane.HTML(f"<b>Key</b>: {tag.key}", styles={'font-size': '16px'}),
                            pn.pane.HTML(f"<b>Value</b>: {tag.value}", styles={'font-size': '16px'}),
                            styles={'margin-top': 'auto', 'margin-bottom': 'auto'}),
                        edit_tag_modal, delete_tag_modal,
                        styles=SECTION_ITEM_STYLES))

        add_button = pn.widgets.Button(name="Add", button_type="success", width=100,
                                       styles={'margin-top': 'auto', 'margin-bottom': 'auto'})
        widgets.append(pn.Row(add_button, styles=SECTION_ITEM_STYLES))
        add_tag_modal = self._add_tag_modal(add_button)

        return pn.Column(
            pn.pane.Markdown("## Pipeline Tags"),
            pn.FlexBox(*widgets),
            add_tag_modal
        )

    def _add_tag_modal(self, trigger):
        """Return modal for adding a new tag to the pipeline"""

        key_input = pn.widgets.TextInput(name="Key", placeholder="Enter tag key")
        value_input = pn.widgets.TextInput(name="Value", placeholder="Enter tag value")

        save_button = pn.widgets.Button(name="Save", button_type="success")
        cancel_button = pn.widgets.Button(name="Cancel")
        modal = pn.Modal("### Add Tag", key_input, value_input, pn.Row(cancel_button, save_button, styles={
            'margin-top': '25px'}))

        def open_modal(event):
            modal.show()

        def save(event):
            new_key = key_input.value.strip()
            new_value = value_input.value.strip()
            if new_key and new_value:
                self.state.tags.append(Tag(key=new_key, value=new_value))
                modal.hide()
                self.view[2] = self._create_tags_section()
            else:
                pn.state.notifications.error("Key or value cannot be empty.", duration=10000)

        def cancel(event):
            key_input.value = ""
            value_input.value = ""
            modal.hide()

        trigger.on_click(open_modal)
        save_button.on_click(save)
        cancel_button.on_click(cancel)

        return modal

    def _tag_edit_modal(self, tag, trigger):
        """Return modal for editing a tag's key and value"""

        key_input = pn.widgets.TextInput(name="Key", value=tag.key)
        value_input = pn.widgets.TextInput(name="Value", value=tag.value)

        save_button = pn.widgets.Button(name="Save", button_type="success")
        cancel_button = pn.widgets.Button(name="Cancel")
        modal = pn.Modal("### Edit Tag", key_input, value_input, pn.Row(cancel_button, save_button, styles={
            'margin-top': '25px'}))

        def open_modal(event):
            modal.show()

        def save(event):
            new_key = key_input.value.strip()
            new_value = value_input.value.strip()
            if new_key and new_value:
                tag.key = new_key
                tag.value = new_value
                modal.hide()
                self.view[2] = self._create_tags_section()
            else:
                pn.state.notifications.error("Key or value cannot be empty.", duration=10000)

        def cancel(event):
            key_input.value = tag.key
            value_input.value = tag.value
            modal.hide()

        trigger.on_click(open_modal)
        save_button.on_click(save)
        cancel_button.on_click(cancel)

        return modal

    def _pipeline_tag_delete_modal(self, tag, trigger):

        confirm_button = pn.widgets.Button(name="Confirm", button_type="danger")
        cancel_button = pn.widgets.Button(name="Cancel")

        modal = pn.Modal("### Delete Tag", "Are you sure you want to delete this tag?",
                         pn.Row(cancel_button, confirm_button))

        def open_modal(event):
            modal.show()

        def confirm(event):
            for i, t in enumerate(self.state.tags):
                if t.key == tag.key and t.value == tag.value:
                    del self.state.tags[i]
                    pn.state.notifications.success(
                        f"Pipeline Tag ({t.key}: {t.value}) deleted successfully.",
                        duration=10000)
                    break
            modal.hide()
            self.view[2] = self._create_tags_section()

        def cancel(event):
            modal.hide()

        trigger.on_click(open_modal)
        confirm_button.on_click(confirm)
        cancel_button.on_click(cancel)

        return modal

    async def clone_pipeline(self, type="staged"):
        """Clones the pipeline and stages or runs it based on the operation type.

        Args:
            type (str): The type of post-cloning operation, either "staged" or "ready".
        """
        try:
            pipeline_input = PipelineInput(
                name=self.state.name,
                parameters=[ParameterInput(name=param.name, value=param.value, type=param.type.name)
                            for param in self.state.parameters] if self.state.parameters else [],
                data_catalog=[
                    DataSetInput(
                        name=ds.name, config=ds.config,
                        tags=[TagInput(key=tag.key, value=tag.value) for tag in ds.tags] if ds.tags else [])
                    for ds in self.state.data_catalog] if self.state.data_catalog else [],
                tags=[TagInput(key=tag.key, value=tag.value) for tag in self.state.tags] if self.state.tags else [],
                runner=self.state.status[-1].runner, state=PipelineInputStatus.STAGED
                if type == "staged" else PipelineInputStatus.READY)
            await self.client.create_pipeline(pipeline_input=pipeline_input)

        except Exception as e:
            pn.state.notifications.error(
                f"Error cloning and {'staging' if type == 'staged' else 'running'} pipeline: {str(e)}", duration=10000)
            return

        pn.state.notifications.success(
            f"Pipeline cloned and {'staged' if type == 'staged' else 'executed'} successfully.", duration=10000)



    def __panel__(self):
        return self._content



from gql import Client, gql
from gql.transport.aiohttp import AIOHTTPTransport
from gql.transport.websockets import WebsocketsTransport
from kedro_graphql.models import PipelineInput, Pipeline, Pipelines, PipelineEvent, PipelineLogMessage, DataSetInput, SignedUrl, SignedUrls
from kedro_graphql.config import load_config
import backoff
from gql.transport.exceptions import TransportQueryError
from typing import Optional, List
import logging

logger = logging.getLogger("kedro-graphql")
CONFIG = load_config()
logger.debug("configuration loaded by {s}".format(s=__name__))

PIPELINE_GQL = """{
                    id
                    parent
                    name
                    describe
                    createdAt
                    projectVersion
                    pipelineVersion
                    kedroGraphqlVersion
                    dataCatalog {
                      name
                      config
                      tags {
                        key
                        value
                      }
                    }
                    nodes {
                      name
                      inputs
                      outputs
                      tags
                    }
                    parameters {
                      name
                      value
                      type
                    }
                    status {
                      session
                      state
                      runner
                      startedAt
                      finishedAt
                      taskId
                      taskName
                      taskArgs
                      taskKwargs
                      taskRequest
                      taskException
                      taskTraceback
                      taskEinfo
                      taskResult
                    }
                    tags {
                      key
                      value
                    }
                  }"""


class KedroGraphqlClient():

    def __init__(self, uri_graphql=None, uri_ws=None, pipeline_gql=None, headers={}, cookies=None):
        """
        Kwargs:
            uri_graphql (str): uri to api [default: http://localhost:5000/graphql]
            uri_ws (str): uri to websocket [default: ws://localhost:5000/graphql]
            pipeline_gql (str): pipeline graphql query [default: kedro_graphql.client.PIPELINE_GQL]

        """
        self.uri_graphql = uri_graphql or CONFIG["KEDRO_GRAPHQL_CLIENT_URI_GRAPHQL"]
        self.uri_ws = uri_ws or CONFIG["KEDRO_GRAPHQL_CLIENT_URI_WS"]
        if cookies:
            self._cookies = "; ".join(
                [f"{key}={value}" for key, value in cookies.items()])
            self._cookies = {"Cookie": self._cookies}
        else:
            self._cookies = {}

        self._headers = headers
        self._headers.update(self._cookies)

        self._aio_transport = AIOHTTPTransport(
            url=self.uri_graphql, headers=headers)

        self._aio_client = Client(transport=self._aio_transport)
        self._aio_session = None
        self.pipeline_gql = pipeline_gql or PIPELINE_GQL

    async def _get_aio_session(self):
        """Get or create an aio session.
        """
        if not self._aio_session:
            logger.info("connecting aio session")
            self._aio_session = await self._aio_client.connect_async(reconnecting=True)
            return self._aio_session
        else:
            return self._aio_session

    async def close_sessions(self):
        """Close any open aio and web sessions.
        """
        if self._aio_session:
            logger.info("closing aio session")
            await self._aio_client.close_async()

    async def execute_query(self, query: str, variable_values: Optional[dict] = None):
        """Make a query to the GraphQL API.

        Kwargs:
            query (str): GraphQL query
            variables (dict): GraphQL variables

        Returns:
            dict: response
        """
        variable_values = variable_values or {}
        session = await self._get_aio_session()
        result = await session.execute(gql(query), variable_values=variable_values)
        return result

    async def create_pipeline(self, pipeline_input: PipelineInput = None, unique_paths: List[str] = None):
        """Create a pipeline

        Kwargs:
            pipeline (PipelineInput): pipeline input object

        Returns:
            Pipeline: pipeline object
        """
        query = """
            mutation createPipeline($pipeline: PipelineInput!, $uniquePaths: [String!]) {
              createPipeline(pipeline: $pipeline, uniquePaths: $uniquePaths) """ + self.pipeline_gql + """
            }
        """

        result = await self.execute_query(query, variable_values={"pipeline": pipeline_input.encode(encoder="graphql"), "uniquePaths": unique_paths})
        return Pipeline.decode(result["createPipeline"], decoder="graphql")

    async def read_pipeline(self, id: str = None):
        """Read a pipeline.
        Kwargs:
            id (str): pipeline id

        Returns:
            Pipeline: pipeline object
        """
        query = """
            query readPipeline($id: String!) {
              readPipeline(id: $id) """ + self.pipeline_gql + """
            }
        """

        result = await self.execute_query(query, variable_values={"id": str(id)})
        return Pipeline.decode(result["readPipeline"], decoder="graphql")

    async def read_pipelines(self, limit: int = 10, cursor: str = None, filter: str = "", sort: str = ""):
        """Read pipelines.

        Kwargs:
            limit (int): limit
            cursor (str): cursor
            filter (str): a valid MongoDb document query filter https://www.mongodb.com/docs/manual/core/document/#std-label-document-query-filter.

        Returns:
            Pipelines (list): an list of pipeline objects
        """
        query = """
            query readPipelines($limit: Int!, $cursor: String, $filter: String, $sort: String) {
              readPipelines(limit: $limit, cursor: $cursor, filter: $filter, sort: $sort) { 
                pageMeta {
                  nextCursor
                }
                pipelines """ + self.pipeline_gql + """
              }
            }
        """

        result = await self.execute_query(query, variable_values={"limit": limit, "cursor": cursor, "filter": filter, "sort": sort})
        return Pipelines.decode(result, decoder="graphql")

    async def update_pipeline(self, id: str = None, pipeline_input: PipelineInput = None, unique_paths: List[str] = None):
        """Update a pipeline

        Kwargs:
            id (str): pipeline id
            pipeline_input (PipelineInput): pipeline input object

        Returns:
            Pipeline: pipeline object
        """
        query = """
            mutation updatePipeline($id: String!, $pipeline: PipelineInput!, $uniquePaths: [String!]) {
              updatePipeline(id: $id, pipeline: $pipeline, uniquePaths: $uniquePaths) """ + self.pipeline_gql + """
            }
        """

        result = await self.execute_query(query, variable_values={"id": str(id), "pipeline": pipeline_input.encode(encoder="graphql"), "uniquePaths": unique_paths})
        return Pipeline.decode(result["updatePipeline"], decoder="graphql")

    async def delete_pipeline(self, id: str = None):
        """Delete a pipeline.

        Kwargs:
            id (str): pipeline id

        Returns:
            Pipeline: pipeline object that was deleted.
        """
        query = """
            mutation deletePipeline($id: String!) {
              deletePipeline(id: $id) """ + self.pipeline_gql + """ 
            }
        """

        result = await self.execute_query(query, variable_values={"id": str(id)})
        return Pipeline.decode(result["deletePipeline"], decoder="graphql")

    async def read_datasets(self, id: str = None, datasets: list[DataSetInput] = None, expires_in_sec: int = 43200):
        """Read a dataset.
        Kwargs:
            id (str): pipeline id
            datasets (list[DataSetInput]): dataset inputs for which to get signed URLs. In order to read specific partitions of a PartitionedDataset, pass a DataSetInput with the dataset name and list of partitions e.g. DataSetInput(name="dataset_name", partitions=["partition1", "partition2"]).
            expires_in_sec (int): number of seconds the signed URL should be valid for

        Returns:
            str: signed URL for reading the dataset
        """
        query = """
            query readDatasets($id: String!, $datasets: [DataSetInput!]!, $expires_in_sec: Int!) {
              readDatasets(id: $id, datasets: $datasets, expiresInSec: $expires_in_sec){
                __typename
                ... on SignedUrl {
                  url
                  file
                  fields {
                    name
                    value
                  }
                }
                ... on SignedUrls {
                  urls {
                    url
                    file
                    fields {
                      name
                      value
                    }
                  }
                }
              }
            }
        """

        result = await self.execute_query(query, variable_values={"id": str(id), "datasets": [d.encode(encoder="graphql") for d in datasets], "expires_in_sec": expires_in_sec})
        urls = []
        for d in result["readDatasets"]:
            if d["__typename"] == "SignedUrl":
                d.pop("__typename")
                urls.append(SignedUrl.decode(d, decoder="graphql"))
            elif d["__typename"] == "SignedUrls":
                d.pop("__typename")
                urls.append(SignedUrls.decode(d, decoder="graphql"))
            else:
                raise TypeError(
                    f"Unexpected type {d['__typename']} returned from readDatasets")

        return urls

    async def create_datasets(self, id: str = None, datasets: list[DataSetInput] = None, expires_in_sec: int = 43200):
        """create a dataset.
        Kwargs:
            id (str): pipeline id
            datasets (list[DataSetInput]): List of datasets for which to create signed URLs. In order to create specific partitions of a PartitionedDataset, pass a DataSetInput with the dataset name and list of partitions e.g. DataSetInput(name="dataset_name", partitions=["partition1", "partition2"]).
            expires_in_sec (int): number of seconds the signed URL should be valid for

        Returns:
            [str]: array of signed URLs for creating the datasets
        """
        query = """
            mutation createDatasets($id: String!, $datasets: [DataSetInput!]!, $expires_in_sec: Int!) {
              createDatasets(id: $id, datasets: $datasets, expiresInSec: $expires_in_sec){
                __typename
                ... on SignedUrl {
                  url
                  file
                  fields {
                    name
                    value
                  }
                }
                ... on SignedUrls {
                  urls {
                    url
                    file
                    fields {
                      name
                      value
                    }
                  }
                }
              }
            }
        """

        result = await self.execute_query(query, variable_values={"id": str(id), "datasets": [d.encode(encoder="graphql") for d in datasets], "expires_in_sec": expires_in_sec})
        urls = []
        for d in result["createDatasets"]:
            if d["__typename"] == "SignedUrl":
                d.pop("__typename")
                urls.append(SignedUrl.decode(d, decoder="graphql"))
            elif d["__typename"] == "SignedUrls":
                d.pop("__typename")
                urls.append(SignedUrls.decode(d, decoder="graphql"))
            else:
                raise TypeError(
                    f"Unexpected type {d['__typename']} returned from createDatasets")

        return urls

    @backoff.on_exception(backoff.expo, Exception, max_time=60, giveup=lambda e: isinstance(e, TransportQueryError))
    async def pipeline_events(self, id: str = None):
        """Subscribe to pipeline events.

        Kwargs:
            id (str): pipeline id

        Returns:
            PipelineEvent (generator): a generator of PipelineEvent objects
        """
        async with Client(
            transport=WebsocketsTransport(url=self.uri_ws, headers=self._headers),
        ) as session:
            query = gql(
                """
                subscription pipelineEvents($id: String!) {
                  pipeline(id: $id) {
                    id
                    taskId
                    status
                    result
                    timestamp
                    traceback
                  }
                }
            """
            )
            logger.info("started pipeline events subscription")
            async for result in session.subscribe(query, variable_values={"id": str(id)}):
                yield PipelineEvent.decode(result, decoder="graphql")

    @backoff.on_exception(backoff.expo, Exception, max_time=60, giveup=lambda e: isinstance(e, TransportQueryError))
    async def pipeline_logs(self, id: str = None):
        """Subscribe to pipeline logs.

        Kwargs:
            id (str): pipeline id

        Returns:
            PipelineLogMessage (generator): a generator of PipelineLogMessage objects
        """
        async with Client(
            transport=WebsocketsTransport(url=self.uri_ws, headers=self._headers),
        ) as session:

            query = gql(
                """
                subscription pipelineLogs($id: String!) {
                  pipelineLogs(id: $id) {
                    id
                    message
                    messageId
                    taskId
                    time
                  }
                }
            """
            )
            logger.info("started pipeline logs subscription")
            async for result in session.subscribe(query, variable_values={"id": str(id)}):
                yield PipelineLogMessage.decode(result, decoder="graphql")



import typing
import strawberry
from starlette.requests import Request
from starlette.websockets import WebSocket
from strawberry.permission import BasePermission
from .logs.logger import logger


def get_permissions(permissions_cls) -> BasePermission:
    module_path, class_name = permissions_cls.rsplit(".", 1)
    module = __import__(module_path, fromlist=[class_name])
    module_class = getattr(module, class_name)
    return module_class


class IsAuthenticatedAction(BasePermission):
    """Base class for authentication permissions using actions.
    """
    message = "User is not authenticated"
    error_extensions = {"code": "UNAUTHORIZED"}

    def __init__(self, action):
        """Initialize the permission with a specific action.
        """
        self._all_actions = ["create_pipeline",
                             "read_pipeline",
                             "read_pipelines",
                             "update_pipeline",
                             "delete_pipeline",
                             "read_pipeline_template",
                             "read_pipeline_templates",
                             "create_dataset",
                             "read_dataset",
                             "subscribe_to_events",
                             "subscribe_to_logs",
                             "create_event"]
        self.action = action

    @staticmethod
    def get_user_info(info: strawberry.Info) -> typing.Optional[typing.Any]:
        """Get user information from the request context.
        This method should be overridden in subclasses if needed.

        Args:
            info: Strawberry Info object containing the request context.

        Returns:
            Optional[Any]: User information, or None if not available.
        """
        raise NotImplementedError(
            "get_user_info method must be implemented in subclasses")

    def has_permission(
        self, source: typing.Any, info: strawberry.Info, **kwargs
    ) -> bool:
        """Check if the user has permission for the specified action.
        This method should be overridden in subclasses.

        Kwargs:
            source: The source of the request, typically the fields of the GraphQL query.
            info: Strawberry Info object containing the request context.
            **kwargs: Additional keyword arguments that may be used in the future.

        Returns:
            bool: True if the user has permission, False otherwise.
        """
        raise NotImplementedError


class IsAuthenticatedAlways(IsAuthenticatedAction):
    """Permission class that always grants access."""

    @staticmethod
    def get_user_info(info: strawberry.Info) -> typing.Optional[typing.Any]:
        """Get user information from the request context.
        This method returns None since this permission always grants access.

        Args:
            info: Strawberry Info object containing the request context.

        Returns:
            Optional[Any]: Always returns None.
        """
        return {"user": None, "email": None, "groups": None}

    def has_permission(
        self, source: typing.Any, info: strawberry.Info, **kwargs
    ) -> bool:
        """Always grants permission regardless of user authentication.

        Kwargs:
            source: The source of the request, typically the fields of the GraphQL query.
            info: Strawberry Info object containing the request context.
            **kwargs: Additional keyword arguments that may be used in the future.

        Returns:
            bool: Always returns True, granting permission.
        """

        logger.info(
            "authentication disabled - permission granted - user=None, action={a}, source={s}".format(s=str(source), a=str(self.action)))
        return True


class IsAuthenticatedXForwardedEmail(IsAuthenticatedAction):
    """Permission class that checks for X-Forwarded-Email header.

    Also supports X-Auth-Request-* headers as fallbacks.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @staticmethod
    def get_user_info(info: strawberry.Info) -> typing.Optional[typing.Any]:
        """Get user information from X-Forwarded-* or X-Auth-Request-* headers.

        Checks for both X-Forwarded-Email/User headers and X-Auth-Request-Email/User headers.

        Args:
            info: Strawberry Info object containing the request context.

        Returns:
            Optional[Any]: Dictionary with 'email' and 'user' keys, or None if not available.
        """
        request: typing.Union[Request, WebSocket] = info.context["request"]

        email = request.headers.get("X-Forwarded-Email") or request.headers.get("x-auth-request-email")
        user = request.headers.get("X-Forwarded-User") or request.headers.get("x-auth-request-user")

        return {"email": email, "user": user}

    def has_permission(
        self, source: typing.Any, info: strawberry.Info, **kwargs
    ) -> bool:
        """Check if the user has permission based on X-Forwarded-Email header.
        Checks for both X-Forwarded-Email and X-Auth-Request-Email headers as fallbacks.
        If either header is present, permission is granted.
        If neither header is present, permission is denied.

        Kwargs:
            source: The source of the request, typically the fields of the GraphQL query.
            info: Strawberry Info object containing the request context.
            **kwargs: Additional keyword arguments that may be used in the future.

        Returns:
            bool: True if the user has permission, False otherwise.
        """
        request: typing.Union[Request, WebSocket] = info.context["request"]

        email = request.headers.get("X-Forwarded-Email") or request.headers.get("x-auth-request-email")
        user = request.headers.get("X-Forwarded-User") or request.headers.get("x-auth-request-user")
        if email:
            logger.info("permission granted - user={u}, action={a}, source={s}".format(
                u=str(email or user),
                a=str(self.action),
                s=str(source)))

            return True
        else:
            logger.info("permission denied - user={u}, action={a}, source={s}".format(
                u=str(user),
                a=str(self.action),
                s=str(source)))

            return False


class IsAuthenticatedXForwardedRBAC(IsAuthenticatedAction):
    """Permission class that checks for X-Forwarded-Groups header and RBAC mapping.

    Also supports X-Auth-Request-Groups headers as fallbacks.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @staticmethod
    def get_user_info(info: strawberry.Info) -> typing.Optional[typing.Any]:
        """Get user information from X-Forwarded-* or X-Auth-Request-* headers.

        Checks for both X-Forwarded-* headers and X-Auth-Request-* headers as fallbacks.

        Args:
            info: Strawberry Info object containing the request context.

        Returns:
            Optional[Any]: Dictionary with 'email', 'groups', and 'user' keys.
        """
        request: typing.Union[Request, WebSocket] = info.context["request"]

        email = request.headers.get("X-Forwarded-Email") or request.headers.get("x-auth-request-email")
        user = request.headers.get("X-Forwarded-User") or request.headers.get("x-auth-request-user")
        groups = request.headers.get("X-Forwarded-Groups") or request.headers.get("x-auth-request-groups")

        return {"email": email,
                "groups": groups,
                "user": user}

    def has_permission(
        self, source: typing.Any, info: strawberry.Info, **kwargs
    ) -> bool:
        """Check if the user has permission based on X-Forwarded-Groups header and RBAC mapping.

        Checks for both X-Forwarded-Groups and X-Auth-Request-Groups headers as fallbacks.
        If the header is present and the user belongs to a group that has the required role for
        the specified action, permission is granted.
        If the header is not present or the user does not belong to a group with the required
        role for the specified action, permission is denied.

        Kwargs:
            source: The source of the request, typically the fields of the GraphQL query.
            info: Strawberry Info object containing the request context.
            **kwargs: Additional keyword arguments that may be used in the future.

        Returns:
            bool: True if the user has permission, False otherwise.
        """
        request: typing.Union[Request, WebSocket] = info.context["request"]

        group_to_role = info.context["request"].app.config.get(
            "KEDRO_GRAPHQL_PERMISSIONS_GROUP_TO_ROLE_MAP", None)

        role_to_action = info.context["request"].app.config.get(
            "KEDRO_GRAPHQL_PERMISSIONS_ROLE_TO_ACTION_MAP", None)

        if not group_to_role:
            logger.warning(
                "KEDRO_GRAPHQL_PERMISSIONS_GROUP_TO_ROLE_MAP is not set in the config, all actions will be denied.")
            return False

        user_groups = request.headers.get("X-Forwarded-Groups", None) or request.headers.get("x-auth-request-groups", None)

        if not user_groups:
            logger.warning(
                "X-Forwarded-Groups header is not set, all actions will be denied.")
            return False
        
        email = request.headers.get("X-Forwarded-Email") or request.headers.get("x-auth-request-email")
        user = request.headers.get("X-Forwarded-User") or request.headers.get("x-auth-request-user")

        for group in user_groups.split(","):
            if group_to_role.get(group, None):
                role = group_to_role[group]
                if role_to_action.get(role, None):
                    if self.action in role_to_action[role]:
                        logger.info("permission granted - user={u}, role={r}, action={a}, source={s}".format(
                            u=str(email or user),
                            r=str(role),
                            a=str(self.action),
                            s=str(source)))
                        return True

        logger.info("permission denied - user:={u}, action={a}, source={s}".format(
            u=str(email or user),
            a=str(self.action),
            s=str(source)))

        return False



import pytest

import json
from pathlib import Path
from urllib.parse import urlparse, parse_qs
import jwt
from ..utilities import kedro_graphql_config
from kedro_graphql.signed_url.local_file_provider import LocalFileProvider
from kedro_graphql.models import DataSet, SignedUrls, SignedUrl


class TestLocalFileProvider:

    config = kedro_graphql_config()

    @pytest.fixture
    def mock_dataset(self, mock_text_in):
        """Fixture for a TextDataset pointing to a test data file"""

        return DataSet(name="test_dataset", config=json.dumps(
            {"type": "text.TextDataset", "filepath": str(mock_text_in)}))

    @pytest.fixture
    def mock_partitioned_dataset(self):
        """Fixture for a PartitionedDataset pointing to test data files"""
        return DataSet(name="test_partitioned_dataset", config=json.dumps(
            {"type": "partitions.PartitionedDataset",
             "path": str(Path("src/tests/data/partitioned_dataset/").resolve()),
             "filename_suffix": ".txt",
             "dataset": {"type": "text.TextDataset"}}
        ))

    def test_create_allowed_local(self, mock_info_context, mock_dataset):
        """Test creating a signed URL for a TextDataset"""

        output = LocalFileProvider.create(
            mock_info_context, mock_dataset, expires_in_sec=10)
        token = output.get_field_value("token")

        payload = jwt.decode(token, self.config["KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_JWT_SECRET_KEY"],
                             algorithms=[self.config["KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_JWT_ALGORITHM"]])

        assert isinstance(payload, dict)

    def test_create_partitioned_local(self, mock_info_context, mock_partitioned_dataset):
        """Test creating signed URLs for specific partitions from a PartitionedDataset"""

        partitions = ["part-0001", "part-0002"]
        output = LocalFileProvider.create(
            mock_info_context, mock_partitioned_dataset, expires_in_sec=10, partitions=partitions)
        assert isinstance(output, SignedUrls)

        expected_files = [mock_partitioned_dataset.parse_config()["path"] + "/" + f + ".txt" for f in
                          partitions]

        for item in output.urls:
            assert isinstance(item, SignedUrl)
            token = item.get_field_value("token")

            payload = jwt.decode(token, self.config["KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_JWT_SECRET_KEY"],
                                 algorithms=[self.config["KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_JWT_ALGORITHM"]])

            assert isinstance(payload, dict)
            assert "filepath" in payload
            assert "exp" in payload
            assert "iat" in payload
            assert payload["filepath"] in expected_files

    def test_signed_url_read_allowed_local(self, mock_info_context, mock_dataset):
        """Test reading a single file from a TextDataset"""

        output = LocalFileProvider.read(
            mock_info_context, mock_dataset, expires_in_sec=10)
        parsed_url = urlparse(output.url)
        query_params = parse_qs(parsed_url.query)
        token = query_params.get("token", [None])[0]

        payload = jwt.decode(token, self.config["KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_JWT_SECRET_KEY"],
                             algorithms=[self.config["KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_JWT_ALGORITHM"]])

        assert isinstance(payload, dict)

        c = json.loads(mock_dataset.config)
        assert Path(payload["filepath"]).resolve() == Path(c["filepath"]).resolve()

    def test_read_partitioned_local_00(self, mock_info_context, mock_partitioned_dataset):
        """Test reading all partitions from a PartitionedDataset"""

        output = LocalFileProvider.read(
            mock_info_context, mock_partitioned_dataset, expires_in_sec=10)
        assert isinstance(output, SignedUrls)

        expected_files = [mock_partitioned_dataset.parse_config()["path"] + "/" + f for f in
                          ["part-0001.txt", "part-0002.txt"]]

        for url in output.urls:
            token = urlparse(url.url).query.split("token=")[1]

            payload = jwt.decode(token, self.config["KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_JWT_SECRET_KEY"],
                                 algorithms=[self.config["KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_JWT_ALGORITHM"]])

            assert isinstance(payload, dict)
            assert "filepath" in payload
            assert "exp" in payload
            assert "iat" in payload
            assert payload["filepath"] in expected_files

    def test_read_partitioned_local_01(self, mock_info_context, mock_partitioned_dataset):
        """Test reading specific partitions from a PartitionedDataset"""

        output = LocalFileProvider.read(
            mock_info_context, mock_partitioned_dataset, expires_in_sec=10, partitions=["part-0001"])

        assert isinstance(output, SignedUrls)

        expected_files = [mock_partitioned_dataset.parse_config()["path"] + "/" + f for f in
                          ["part-0001.txt"]]

        for url in output.urls:
            token = urlparse(url.url).query.split("token=")[1]

            payload = jwt.decode(token, self.config["KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_JWT_SECRET_KEY"],
                                 algorithms=[self.config["KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_JWT_ALGORITHM"]])

            assert isinstance(payload, dict)
            assert "filepath" in payload
            assert "exp" in payload
            assert "iat" in payload
            assert payload["filepath"] in expected_files



"""A python panel app for visualizing Kedro pipelines."""
import panel as pn
from panel.custom import JSComponent
import param
from kedro_graphql.client import KedroGraphqlClient


# pn.extension(theme="dark")
pn.extension("modal")
pn.extension(
    disconnect_notification='Connection lost, try reloading the page!',
    ready_notification='Application fully loaded.',
    template='bootstrap'
)


class NavigationSidebarButton(pn.viewable.Viewer):
    """A button for navigating the sidebar in the Kedro GraphQL UI.
    This button is used to navigate to different pages in the UI, such as pipelines, nodes, or data catalog.
    It updates the URL to reflect the current page and changes its appearance based on whether it is the active page.

    Attributes:
        page (str): The name of the page this button navigates to.
        name (str): The display name of the button.
        spec (dict): The specification for the UI, including configuration and pages.
    """
    page = param.String(default="pipelines")
    name = param.String(default="pipelines")
    spec = param.Dict(default={})

    def navigate(self, event):
        """Navigate to the specified page."""
        pn.state.location.pathname = self.spec["panel_get_server_kwargs"]["prefix"]
        pn.state.location.search = "?page=" + self.name.lower()

    async def build_button(self):
        """Builds the navigation button for the sidebar.
        This button will change its appearance based on whether it is the current page.

        Returns:
            pn.widgets.Button: A button that navigates to the specified page.
        """
        if self.page == self.name.lower():
            b0_type = "primary"
            b0_icon = "chevron-right-pipe"
        else:
            b0_type = "default"
            b0_icon = None

        button = pn.widgets.Button(
            name=self.name,
            button_type=b0_type,
            sizing_mode="stretch_width",
            icon=b0_icon
        )
        pn.bind(self.navigate, button, watch=True)
        return button

    def __panel__(self):
        return self.build_button


class TemplateMainFactory(pn.viewable.Viewer):
    """A factory for building the main content of the Kedro GraphQL UI template.
    This component dynamically builds the main content based on the current page specified in the URL.
    It uses the `spec` dictionary to determine which module to load for each page.

    Attributes:
        page (str): The current page to display in the main content.
        spec (dict): The specification for the UI, including configuration and pages.
    """
    page = param.String(default="pipelines")
    spec = param.Dict(default={})

    def __init__(self, **params):
        super().__init__(**params)

    @param.depends("page", "spec")
    async def build_page(self):
        """Builds the main content of the template based on the current page."""
        yield pn.indicators.LoadingSpinner(value=True, width=25, height=25)
        if self.spec["pages"][self.page].get("params", False):
            params = self.spec["pages"][self.page]["params"]
        else:
            params = {}
        yield self.spec["pages"][self.page]["module"](spec=self.spec, **params)

    def __panel__(self):
        pn.state.location.sync(
            self, {"page": "page"})

        return pn.Column(self.build_page)


class KedroGraphqlMaterialTemplate(pn.template.MaterialTemplate):
    """A Material Design template for the Kedro GraphQL UI.
    This template includes a sidebar for navigation and a main content area that
    displays different pages based on the current URL.
    It uses the `spec` dictionary to configure the sidebar and main content.

    Attributes:
        title (str): The title of the template.
        sidebar_width (int): The width of the sidebar in pixels.
        page (str): The current page to display in the main content.
        spec (dict): The specification for the UI, including configuration and pages.
    """

    page = param.String(default="pipelines")
    spec = param.Dict(default={})

    def user_menu_action(self, event, user_info_modal, login_link, logout_link):
        """ Handles user menu actions such as showing user information, logging in, or logging out.
        Args:
            event (str): The action triggered by the user menu.
            user_info_modal (pn.Modal): The modal to display user information.
            login_link (str): The URL for the login endpoint.
            logout_link (str): The URL for the logout endpoint.
        """
        if event == "User Information":
            user_info_modal.show()
        elif event == "Login":
            pn.state.location.pathname = login_link
            pn.state.location.search = ""
            pn.state.location.reload = True
        elif event == "Logout":
            pn.state.location.pathname = logout_link
            pn.state.location.search = ""
            pn.state.location.reload = True

    async def build_user_menu(self):
        """Asynchronously retrieves the user context for the template.
        This method can be overridden to provide custom user context data.
        """
        # if hasattr(pn.state, "user_info") and pn.state.user_info:
        # user = pn.state.user_info.get("name", False) or pn.state.user or "Guest"

        user = pn.state.headers.get("X-Forwarded-User", None) or pn.state.headers.get("X-Auth-Request-User", None)

        if user:
            email = pn.state.headers.get("X-Forwarded-Email", None) or pn.state.headers.get("X-Auth-Request-Email", None)
            groups = pn.state.headers.get("X-Forwarded-Groups", None) or pn.state.headers.get("X-Auth-Request-Groups", None)
            user_info = {"user": user, "email": email, "groups": groups}
        else:
            user = "Guest"
            user_info = {"user": "Guest", "email": "Guest", "groups": None}

        user_menu_items = [("User Information", "User Information")]

        login_link = self.spec["panel_get_server_kwargs"].get(
            "login_endpoint", "/login")
        logout_link = self.spec["panel_get_server_kwargs"].get(
            "logout_endpoint", "/logout")

        if user == "Guest":
            if pn.config.oauth_provider:
                user_menu_items.append("Login")
            elif pn.config.basic_auth:
                user_menu_items.append("Login")
        else:
            user_menu_items.append("Logout")

        user_menu = pn.widgets.MenuButton(name=user_info["email"].split("@")[0], icon="circle-letter-" + user_info["email"][0].lower(
        ), icon_size="2em", items=user_menu_items, sizing_mode="stretch_width",)
        user_info_modal = pn.Modal(user_info, name='User Information',
                                   margin=20, background_close=True, show_close_button=True)
        pn.bind(self.user_menu_action, user_menu.param.clicked,
                user_info_modal, login_link, logout_link, watch=True)
        yield pn.Column(pn.Row(user_menu), pn.Row(user_info_modal, height=0)
                        )

    def init_client(self, spec):
        """Initializes the Kedro GraphQL client with the provided specification.
        This method sets up the client to connect to the GraphQL API and WebSocket.

        Args:
            spec (dict): The specification for the UI, including configuration and pages.
        """

        if pn.state.cookies:
            cookies = pn.state.cookies
        else:
            cookies = None

        client = KedroGraphqlClient(
            uri_graphql=spec["config"]["client_uri_graphql"],
            uri_ws=spec["config"]["client_uri_ws"],
            cookies=cookies
        )

        spec["config"]["client"] = client
        return spec

    def __init__(self, title="kedro-graphql", spec=None):
        """Initializes the KedroGraphqlMaterialTemplate with a title and specification.
        Args:
            title (str): The title of the template.
            spec (dict): The specification for the UI, including configuration and pages.
        """
        super().__init__(
            title=spec["panel_get_server_kwargs"]["title"],
            sidebar_width=200)

        self.spec = self.init_client(spec)
        self.sidebar.append(self.build_user_menu)
        self.sidebar.append(pn.layout.Divider())
        for nav in spec["nav"]["sidebar"]:
            next_button = NavigationSidebarButton(name=nav["name"], spec=spec)
            pn.state.location.sync(next_button, {"page": "page"})
            self.sidebar.append(next_button)

        self.main.append(pn.Column(TemplateMainFactory(spec=spec)))
        # pn.state.location.pathname = spec["panel_get_server_kwargs"]["prefix"]
        pn.state.location.sync(
            self, {"page": "page"})


class NavigationSidebarButtonV2(pn.viewable.Viewer):
    """A button for navigating the sidebar in the Kedro GraphQL UI.
    This button is used to navigate to different pages in the UI, such as pipelines, nodes, or data catalog.
    It updates the URL to reflect the current page and changes its appearance based on whether it is the active page.

    Attributes:
        page (str): The name of the page this button navigates to.
        name (str): The display name of the button.
        spec (dict): The specification for the UI, including configuration and pages.
    """
    page = param.String(default="pipelines")
    name = param.String(default="pipelines")
    spec = param.Dict(default={})

    def navigate(self, event):
        """Navigate to the specified page."""
        pn.state.location.pathname = self.spec["panel_get_server_kwargs"]["prefix"] + self.name.lower()
        pn.state.location.search = ""
        pn.state.location.reload = True

    async def build_button(self):
        """Builds the navigation button for the sidebar.
        This button will change its appearance based on whether it is the current page.

        Returns:
            pn.widgets.Button: A button that navigates to the specified page.
        """
        if pn.state.location.pathname == "/" + self.name.lower():
            b0_type = "primary"
            b0_icon = "chevron-right-pipe"
        else:
            b0_type = "default"
            b0_icon = None

        button = pn.widgets.Button(
            name=self.name,
            button_type=b0_type,
            sizing_mode="stretch_width",
            icon=b0_icon
        )
        pn.bind(self.navigate, button, watch=True)
        yield button

    def __panel__(self):
        return self.build_button


class KedroGraphqlMaterialTemplateV2(pn.template.MaterialTemplate):
    """A Material Design template for the Kedro GraphQL UI.
    This template includes a sidebar for navigation and a main content area that
    displays different pages based on the current URL.
    It uses the `spec` dictionary to configure the sidebar and main content.

    Attributes:
        title (str): The title of the template.
        sidebar_width (int): The width of the sidebar in pixels.
        page (str): The current page to display in the main content.
        spec (dict): The specification for the UI, including configuration and pages.
    """
    spec = param.Dict(default={})

    def user_menu_action(self, event, user_info_modal, login_link, logout_link):
        """ Handles user menu actions such as showing user information, logging in, or logging out.
        Args:
            event (str): The action triggered by the user menu.
            user_info_modal (pn.Modal): The modal to display user information.
            login_link (str): The URL for the login endpoint.
            logout_link (str): The URL for the logout endpoint.
        """
        if event == "User Information":
            user_info_modal.show()
        elif event == "Login":
            pn.state.location.pathname = login_link
            pn.state.location.search = ""
            pn.state.location.reload = True
        elif event == "Logout":
            pn.state.location.pathname = logout_link
            pn.state.location.search = ""
            pn.state.location.reload = True

    async def build_user_menu(self):
        """Asynchronously retrieves the user context for the template.
        This method can be overridden to provide custom user context data.
        """
        
        user = pn.state.headers.get("X-Forwarded-User", None) or pn.state.headers.get("X-Auth-Request-User", None)

        if user:
            email = pn.state.headers.get("X-Forwarded-Email", None) or pn.state.headers.get("X-Auth-Request-Email", None)
            groups = pn.state.headers.get("X-Forwarded-Groups", None) or pn.state.headers.get("X-Auth-Request-Groups", None)
            user_info = {"user": user, "email": email, "groups": groups}
        else:
            user = "Guest"
            user_info = {"user": "Guest", "email": "Guest", "groups": None}

        user_menu_items = [("User Information", "User Information")]

        login_link = self.spec["panel_get_server_kwargs"].get(
            "login_endpoint", "/login")
        logout_link = self.spec["panel_get_server_kwargs"].get(
            "logout_endpoint", "/logout")

        if user == "Guest":
            if pn.config.oauth_provider:
                user_menu_items.append("Login")
            elif pn.config.basic_auth:
                user_menu_items.append("Login")
        else:
            user_menu_items.append("Logout")

        user_menu = pn.widgets.MenuButton(name=user_info["email"].split("@")[0], icon="circle-letter-" + user_info["email"][0].lower(
        ), icon_size="2em", items=user_menu_items, sizing_mode="stretch_width",)
        user_info_modal = pn.Modal(user_info, name='User Information',
                                   margin=20, background_close=True, show_close_button=True)
        pn.bind(self.user_menu_action, user_menu.param.clicked,
                user_info_modal, login_link, logout_link, watch=True)
        yield pn.Column(pn.Row(user_menu), pn.Row(user_info_modal, height=0)
                        )

    def init_client(self, spec):
        """Initializes the Kedro GraphQL client with the provided specification.
        This method sets up the client to connect to the GraphQL API and WebSocket.

        Args:
            spec (dict): The specification for the UI, including configuration and pages.
        """

        if pn.state.cookies:
            cookies = pn.state.cookies
        else:
            cookies = None

        client = KedroGraphqlClient(
            uri_graphql=spec["config"]["client_uri_graphql"],
            uri_ws=spec["config"]["client_uri_ws"],
            cookies=cookies
        )

        spec["config"]["client"] = client
        return spec

    def __init__(self, title="kedro-graphql", spec=None, main=[]):
        """Initializes the KedroGraphqlMaterialTemplate with a title and specification.
        Args:
            title (str): The title of the template.
            spec (dict): The specification for the UI, including configuration and pages.
        """
        super().__init__(
            title=spec["panel_get_server_kwargs"]["title"],
            base_url=spec["panel_get_server_kwargs"]["prefix"],
            sidebar_width=200)

        self.spec = self.init_client(spec)
        self.sidebar.append(self.build_user_menu)
        self.sidebar.append(pn.layout.Divider())
        for nav in spec["nav"]["sidebar"]:
            next_button = NavigationSidebarButtonV2(name=nav["name"], spec=spec)
            self.sidebar.append(next_button)

        # fix for browser navigation with forward and back buttons
        fix = pn.pane.HTML("""
        <script>
        window.addEventListener('popstate', function(event) {
          window.location.replace(window.location.href);
        });
        </script>
        """)

        self.main.append(fix)
        self.main.append(main)



from kedro_graphql.config import load_config


def kedro_graphql_config():

    config = load_config()

    # enable events endpoint
    config["KEDRO_GRAPHQL_EVENTS_CONFIG"] = {"event00": {
        "source": "example.com", "type": "com.example.event"}}

    # use "test_pipelines" as the collection name for testing
    config["KEDRO_GRAPHQL_MONGO_DB_COLLECTION"] = "test_pipelines"
    config["KEDRO_GRAPHQL_MONGO_DB_NAME"] = "test_pipelines"

    return config



import asyncio
from base64 import b64decode, b64encode
from datetime import datetime
from importlib import import_module
from typing import Optional, Union, List, Callable, Any
import json
from collections.abc import AsyncGenerator, Iterable
from graphql.execution import ExecutionContext as GraphQLExecutionContext

import strawberry
from bson.objectid import ObjectId
from celery.states import UNREADY_STATES, READY_STATES
from fastapi.encoders import jsonable_encoder
from kedro.framework.project import pipelines
from starlette.concurrency import run_in_threadpool
from strawberry.extensions import SchemaExtension
from strawberry.permission import PermissionExtension
from strawberry.tools import merge_types
from strawberry.types import Info
from strawberry.directive import StrawberryDirective
from strawberry.types.base import StrawberryType
from strawberry.types.scalar import ScalarDefinition, ScalarWrapper
from strawberry.schema.config import StrawberryConfig
from strawberry.scalars import JSON
from strawberry.extensions import FieldExtension

from . import __version__ as kedro_graphql_version
from .config import load_config
from .pipeline_event_monitor import PipelineEventMonitor
from .hooks import InvalidPipeline
from .logs.logger import PipelineLogStream, logger
from .models import (
    DataSetInput,
    PageMeta,
    Pipeline,
    PipelineEvent,
    PipelineInput,
    PipelineLogMessage,
    Pipelines,
    PipelineStatus,
    PipelineTemplate,
    PipelineTemplates,
    SignedUrl,
    SignedUrls,
    State,
    DataSetInput
)
from .tasks import run_pipeline
from .permissions import get_permissions
from .signed_url.base import SignedUrlProvider
from .utils import generate_unique_paths

CONFIG = load_config()
logger.debug("configuration loaded by {s}".format(s=__name__))

PERMISSIONS_CLASS = get_permissions(CONFIG.get("KEDRO_GRAPHQL_PERMISSIONS"))
logger.info("{s} using permissions class: {d}".format(s=__name__, d=PERMISSIONS_CLASS))


def encode_cursor(id: int) -> str:
    """
    Encodes the given id into a cursor.

    :param id: The ID to encode.

    :return: The encoded cursor.
    """
    return b64encode(f"cursor:{id}".encode("ascii")).decode("ascii")


def decode_cursor(cursor: str) -> int:
    """
    Decodes the ID from the given cursor.

    :param cursor: The cursor to decode.

    :return: The decoded user ID.
    """
    cursor_data = b64decode(cursor.encode("ascii")).decode("ascii")
    return cursor_data.split(":")[1]


class DataSetConfigException(Exception):
    """``DataSetConfigException`` raised by ``DataSetConfigExtension`` implementations
    in case of failure.

    ``DataSetConfigExtensions`` implementations should provide instructive
    information in case of failure.
    """

    pass


class PipelineExtension(FieldExtension):
    """
    Intercepts Pipeline arguments to mask filepaths before returning.
    This extension should be added to any Query or Mutation that returns a Pipeline or Pipelines object.
    This extension requires the KEDRO_GRAPHQL_DATASET_FILEPATH_MASKS config to be set to have any effect.
    """

    def resolve(
        self, next_: Callable[..., Any], source: Any, info: strawberry.Info, **kwargs
    ):
        """
        Intercepts Pipeline and Pipelines return values to mask filepaths before returning.

        Args:
            next_ (Callable[..., Any]): The next resolver in the chain.
            source (Any): The source object.
            info (strawberry.Info): The GraphQL execution context.
            **kwargs: Additional keyword arguments.

        Returns:
            Union[Pipeline, Pipelines]: The Pipeline or Pipelines object with masked filepaths.
        """
        # call original resolver
        pipeline = next_(source, info, **kwargs)

        if isinstance(pipeline, Pipeline):
            # mask filepaths before returning
            return PipelineSanitizer.mask_filepaths(
                pipeline, CONFIG["KEDRO_GRAPHQL_DATASET_FILEPATH_MASKS"])
        elif isinstance(pipeline, Pipelines):
            pipelines = []
            for p in pipeline.pipelines:
                pipelines.append(PipelineSanitizer.mask_filepaths(
                    p, CONFIG["KEDRO_GRAPHQL_DATASET_FILEPATH_MASKS"]))
            pipeline.pipelines = pipelines
            return pipeline

    async def resolve_async(
        self, next_: Callable[..., Any], source: Any, info: strawberry.Info, **kwargs
    ):
        """
        Async version of resolve for async resolvers.
        Intercepts Pipeline and Pipelines return values to mask filepaths before returning.

        Args:
            next_ (Callable[..., Any]): The next resolver in the chain.
            source (Any): The source object.
            info (strawberry.Info): The GraphQL execution context.
            **kwargs: Additional keyword arguments.

        Returns:
            Union[Pipeline, Pipelines]: The Pipeline or Pipelines object with masked filepaths.
        """
        # call original resolver
        pipeline = await next_(source, info, **kwargs)

        if isinstance(pipeline, Pipeline):
            # mask filepaths before returning
            return PipelineSanitizer.mask_filepaths(
                pipeline, CONFIG["KEDRO_GRAPHQL_DATASET_FILEPATH_MASKS"])
        elif isinstance(pipeline, Pipelines):
            pipelines = []
            for p in pipeline.pipelines:
                pipelines.append(PipelineSanitizer.mask_filepaths(
                    p, CONFIG["KEDRO_GRAPHQL_DATASET_FILEPATH_MASKS"]))
            pipeline.pipelines = pipelines
            return pipeline


class PipelineInputExtension(FieldExtension):
    """
    Intercepts PipelineInput arguments to unmask and validate filepaths before
    passing to the resolver, then masks filepaths again before returning the Pipeline result.
    This extension should be added to any Mutation that takes a PipelineInput argument.
    This extension requires the KEDRO_GRAPHQL_DATASET_FILEPATH_MASKS and/or KEDRO_GRAPHQL_DATASET_FILEPATH_ALLOWED_ROOTS
    config to be set to have any effect.
    """

    def resolve(
        self, next_: Callable[..., Any], source: Any, info: strawberry.Info, **kwargs
    ):
        """
        Intercepts PipelineInput argument to unmask and validate filepaths before
        passing to the resolver, then masks filepaths again before returning the Pipeline result.

        Args:
            next_ (Callable[..., Any]): The next resolver in the chain.
            source (Any): The source object.
            info (strawberry.Info): The GraphQL execution context.
            **kwargs: Additional keyword arguments.

        Returns:
            Pipeline: The Pipeline object with masked filepaths.

        Raises:
            DataSetConfigException: If any dataset filepath does not start with allowed prefixes.
        """
        # intercept PipelineInput and sanitize filepaths
        pipeline_input = kwargs["pipeline"]

        kwargs["pipeline"] = PipelineSanitizer.unmask_filepaths(
            pipeline_input, CONFIG["KEDRO_GRAPHQL_DATASET_FILEPATH_MASKS"])

        PipelineSanitizer.sanitize_filepaths(
            pipeline_input, CONFIG["KEDRO_GRAPHQL_DATASET_FILEPATH_ALLOWED_ROOTS"])

        # call original resolver
        pipeline = next_(source, info, **kwargs)
        # mask filepaths again before returning
        return PipelineSanitizer.mask_filepaths(
            pipeline, CONFIG["KEDRO_GRAPHQL_DATASET_FILEPATH_MASKS"])

    async def resolve_async(
        self, next_: Callable[..., Any], source: Any, info: strawberry.Info, **kwargs
    ):
        """
        Async version of resolve for async resolvers.
        Intercepts PipelineInput argument to unmask and validate filepaths before
        passing to the resolver, then masks filepaths again before returning the Pipeline result.

        Args:
            next_ (Callable[..., Any]): The next resolver in the chain.
            source (Any): The source object.
            info (strawberry.Info): The GraphQL execution context.
            **kwargs: Additional keyword arguments.

        Returns:
            Pipeline: The Pipeline object with masked filepaths.

        Raises:
            DataSetConfigException: If any dataset filepath does not start with allowed prefixes.
        """
        # intercept PipelineInput and sanitize filepaths
        pipeline_input = kwargs["pipeline"]

        kwargs["pipeline"] = PipelineSanitizer.unmask_filepaths(
            pipeline_input, CONFIG["KEDRO_GRAPHQL_DATASET_FILEPATH_MASKS"])

        PipelineSanitizer.sanitize_filepaths(
            pipeline_input, CONFIG["KEDRO_GRAPHQL_DATASET_FILEPATH_ALLOWED_ROOTS"])

        # call original resolver
        pipeline = await next_(source, info, **kwargs)
        # mask filepaths again before returning
        return PipelineSanitizer.mask_filepaths(
            pipeline, CONFIG["KEDRO_GRAPHQL_DATASET_FILEPATH_MASKS"])


class PipelineSanitizer:

    @staticmethod
    def sanitize_filepaths(pipeline: Pipeline | PipelineInput, allowed_roots: list[str]) -> None:
        """Raises DataSetConfigException if any dataset filepath does not start with any of the allowed roots.

        Args:
            pipeline (Pipeline | PipelineInput): The pipeline to sanitize.
            allowed_roots (list[str]): The list of allowed roots.
        Returns:
            None

        Raises:
            DataSetConfigException: If any dataset filepath does not start with allowed prefixes.
        """
        if pipeline.data_catalog:
            for d in pipeline.data_catalog:
                c = json.loads(d.config)
                if c.get("filepath"):
                    if len(allowed_roots) > 0 and not any(c["filepath"].startswith(root) for root in allowed_roots):
                        raise DataSetConfigException(
                            "filepath " + c["filepath"] + " not allowed")

    @classmethod
    def mask_filepaths(cls, pipeline: Pipeline | PipelineInput, masks: list[dict]) -> Pipeline | PipelineInput:
        """
        Masks filepaths in the pipeline's data catalog according to the given masks.

        Args:
            pipeline (Pipeline | PipelineInput): The pipeline to mask.
            masks (list[dict]): The list of masks.

        Returns:
            Pipeline | PipelineInput: The pipeline with masked filepaths.
        """
        if pipeline.data_catalog:
            for d in pipeline.data_catalog:
                try:
                    c = json.loads(d.config)
                    if c.get("filepath"):
                        for mask in masks:
                            if c["filepath"].startswith(mask["prefix"]):
                                c["filepath"] = c["filepath"].replace(
                                    mask["prefix"], mask["mask"])
                        d.config = json.dumps(c)

                except Exception as e:
                    logger.warning(
                        f"Could not parse config for dataset {d.name}: {e}")
        return pipeline

    @classmethod
    def unmask_filepaths(cls, pipeline: Pipeline | PipelineInput, masks: list[dict]) -> Pipeline | PipelineInput:
        """
        Unmasks filepaths in the pipeline's data catalog according to the given masks.

        Args:
            pipeline (Pipeline | PipelineInput): The pipeline to unmask.
            masks (list[dict]): The list of masks.

        Returns:
            Pipeline | PipelineInput: The pipeline with unmasked filepaths.
        """
        if pipeline.data_catalog:
            for d in pipeline.data_catalog:
                try:
                    c = json.loads(d.config)
                    if c.get("filepath"):
                        for mask in masks:
                            if c["filepath"].startswith(mask["mask"]):
                                c["filepath"] = c["filepath"].replace(
                                    mask["mask"], mask["prefix"])
                        d.config = json.dumps(c)
                except Exception as e:
                    logger.warning(
                        f"Could not parse config for dataset {d.name}: {e}")
        return pipeline


@strawberry.type
class Query:
    @strawberry.field(description="Get a pipeline template.", extensions=[PermissionExtension(permissions=[PERMISSIONS_CLASS(action="read_pipeline_template")])])
    def pipeline_template(self, info: Info, id: str) -> PipelineTemplate:
        for p in info.context["request"].app.kedro_pipelines_index:
            if p.id == id:
                logger.info(
                    f"user={PERMISSIONS_CLASS.get_user_info(info)['email']}, action=read_pipeline_template, id={id}")
                return p

    @strawberry.field(description="Get a list of pipeline templates.", extensions=[PermissionExtension(permissions=[PERMISSIONS_CLASS(action="read_pipeline_templates")])])
    def pipeline_templates(self, info: Info, limit: int, cursor: Optional[str] = None) -> PipelineTemplates:
        if cursor is not None:
            # decode the user ID from the given cursor.
            pipe_id = ObjectId(decode_cursor(cursor=cursor))
        else:
            # unix epoch Jan 1, 1970 as objectId
            pipe_id = ObjectId("100000000000000000000000")

        # filter the pipeline template data, going through the next set of results.
        filtered_data = [pipe for pipe in info.context["request"].app.kedro_pipelines_index
                         if pipe.id.generation_time >= pipe_id.generation_time]

        # slice the relevant pipeline template data (Here, we also slice an
        # additional pipe instance, to prepare the next cursor).
        sliced_pipes = filtered_data[: limit + 1]

        if len(sliced_pipes) > limit:
            # calculate the client's next cursor.
            last_pipe = sliced_pipes.pop(-1)
            next_cursor = encode_cursor(id=last_pipe.id)
        else:
            # We have reached the last page, and
            # don't have the next cursor.
            next_cursor = None
        logger.info(
            f"user={PERMISSIONS_CLASS.get_user_info(info)['email']}, action=read_pipeline_templates, limit={limit}, cursor={cursor}")
        return PipelineTemplates(
            pipeline_templates=sliced_pipes, page_meta=PageMeta(
                next_cursor=next_cursor)
        )

    @strawberry.field(description="Get a pipeline instance.", extensions=[PermissionExtension(permissions=[PERMISSIONS_CLASS(action="read_pipeline")]), PipelineExtension()])
    async def read_pipeline(self, id: str, info: Info) -> Pipeline:
        try:
            # Run blocking database read in thread pool
            p = await run_in_threadpool(info.context["request"].app.backend.read, id=id)
            if p is None:
                raise InvalidPipeline(
                    f"Pipeline {id} does not exist in the project.")
        except Exception as e:
            raise InvalidPipeline(f"Error retrieving pipeline {id}: {e}")
        logger.info(
            f"user={PERMISSIONS_CLASS.get_user_info(info)['email']}, action=read_pipeline, id={id}")
        return p

    @strawberry.field(description="Get a list of pipeline instances.", extensions=[PermissionExtension(permissions=[PERMISSIONS_CLASS(action="read_pipelines")]), PipelineExtension()])
    async def read_pipelines(self, info: Info, limit: int, cursor: Optional[str] = None, filter: Optional[str] = "",
                       sort: Optional[str] = "") -> Pipelines:
        if cursor is not None:
            # decode the user ID from the given cursor.
            pipe_id = decode_cursor(cursor=cursor)
        else:
            pipe_id = "000000000000000000000000"  # unix epoch Jan 1, 1970 as objectId

        # Run blocking database list in thread pool
        results = await run_in_threadpool(
            info.context["request"].app.backend.list,
            cursor=pipe_id, limit=limit + 1, filter=filter, sort=sort)
        if len(results) > limit:

            # calculate the client's next cursor.
            last_pipe = results.pop(-1)
            next_cursor = encode_cursor(id=last_pipe.id)
        else:
            # We have reached the last page, and
            # don't have the next cursor.
            next_cursor = None

        logger.info(
            f"user={PERMISSIONS_CLASS.get_user_info(info)['email']}, action=read_pipelines, filter={filter}, limit={limit}, sort={sort}, cursor={cursor}")
        return Pipelines(
            pipelines=results, page_meta=PageMeta(next_cursor=next_cursor)
        )

    @strawberry.field(description="Read a dataset with a signed URL", extensions=[PermissionExtension(permissions=[PERMISSIONS_CLASS(action="read_dataset")])])
    async def read_datasets(self, id: str, info: Info, datasets: List[DataSetInput], expires_in_sec: int = CONFIG["KEDRO_GRAPHQL_SIGNED_URL_MAX_EXPIRES_IN_SEC"]) -> List[SignedUrl | SignedUrls | None]:
        """
        Get a signed URL for downloading a dataset.

        Args:
            id (str): The ID of the pipeline.
            info (Info): The GraphQL execution context.
            datasets (List[DataSetInput]): The datasets to get signed URLs for. In order to read specific partitions of a PartitionedDataset, pass a DataSetInput with the dataset name and list of partitions e.g. DataSetInput(name="dataset_name", partitions=["partition1", "partition2"]).
            expires_in_sec (int): The number of seconds the signed URL should be valid for.
        Returns:
            List[SignedUrl | SignedUrls | None]: An array of signed URLs for downloading the dataset or None if not applicable.

        Raises:
            ValueError: If expires_in_sec is greater than max expires_in_sec
            DataSetConfigError: If the dataset configuration is invalid or cannot be parsed.
            TypeError: If the signed URL provider does not inherit from SignedUrlProvider.
        """

        if expires_in_sec > CONFIG["KEDRO_GRAPHQL_SIGNED_URL_MAX_EXPIRES_IN_SEC"]:
            raise ValueError(
                f"expires_in_sec cannot be greater than {CONFIG['KEDRO_GRAPHQL_SIGNED_URL_MAX_EXPIRES_IN_SEC']} seconds ({CONFIG['KEDRO_GRAPHQL_SIGNED_URL_MAX_EXPIRES_IN_SEC'] // 3600} hours)")

        urls = []
        # Run blocking database read in thread pool
        p = await run_in_threadpool(info.context["request"].app.backend.read, id=id)

        catalog = {d.name: d for d in p.data_catalog}

        for d in datasets:
            dataset = catalog.get(d.name, None)
            if dataset is None:
                logger.warning(
                    f"Dataset '{d.name}' not found in the data catalog of pipeline_name={p.name} pipeline_id={p.id}. SignedURL set to None.")
                urls.append(None)
                continue

            module_path, class_name = CONFIG["KEDRO_GRAPHQL_SIGNED_URL_PROVIDER"].rsplit(
                ".", 1)
            module = import_module(module_path)
            cls = getattr(module, class_name)

            if not issubclass(cls, SignedUrlProvider):
                raise TypeError(
                    f"{class_name} must inherit from SignedUrlProvider")

            logger.info(
                f"user={PERMISSIONS_CLASS.get_user_info(info)['email']}, action=read_dataset, dataset={dataset.name}, expires_in_sec={expires_in_sec}")
            urls.append(cls.read(info, dataset, expires_in_sec, d.partitions))

        return urls


@strawberry.type
class Mutation:
    @strawberry.mutation(description="Execute a pipeline.", extensions=[PermissionExtension(permissions=[PERMISSIONS_CLASS(action="create_pipeline")]), PipelineInputExtension()])
    async def create_pipeline(self, pipeline: PipelineInput, info: Info, unique_paths: Optional[List[str]] = None) -> Pipeline:
        """
        - is validation against template needed, e.g. check DataSet type or at least check dataset names
        """

        if pipeline.name not in pipelines.keys():
            raise InvalidPipeline(
                f"Pipeline {pipeline.name} does not exist in the project.")

        d = jsonable_encoder(pipeline)
        p = Pipeline.decode(d)
        p.describe = info.context["request"].app.kedro_pipelines[p.name].describe()
        p.nodes = info.context["request"].app.kedro_pipelines[p.name].nodes
        serial = p.encode(encoder="kedro")

        runner = d.get(
            "runner") or info.context["request"].app.config["KEDRO_GRAPHQL_RUNNER"]
        # credentials not supported yet
        # merge any credentials with inputs and outputs
        # credentials are intentionally not persisted
        # NOTE celery result may persist creds in task result?

        started_at = datetime.now()
        p.created_at = started_at

        # Get kedro project, kedro-graphql, and pipeline versions
        p.project_version = CONFIG.get("KEDRO_PROJECT_VERSION", None)
        p.kedro_graphql_version = kedro_graphql_version
        p.pipeline_version = None
        package_name = CONFIG.get("KEDRO_PROJECT_NAME", None)
        if package_name:
            try:
                module = import_module(
                    f".pipelines.{pipeline.name}", package=package_name)
                p.pipeline_version = getattr(module, "__version__", None)
            except Exception as e:
                logger.info(f"Could not find pipeline version: {e}")

        if d["state"] == "STAGED":
            p.status.append(PipelineStatus(state=State.STAGED,
                                           runner=runner,
                                           session=None,
                                           started_at=None,
                                           finished_at=None,
                                           task_id=None,
                                           task_name=None))
            logger.info(f'Staging pipeline {p.name}')
            # Run blocking database create in thread pool
            p = await run_in_threadpool(info.context["request"].app.backend.create, p)
            if unique_paths:
                p = generate_unique_paths(p, unique_paths)
                # Run blocking database update in thread pool
                p = await run_in_threadpool(info.context["request"].app.backend.update, p)

            logger.info(
                f"user={PERMISSIONS_CLASS.get_user_info(info)['email']}, action=create_pipeline, id={p.id}, name={p.name}, state=STAGED")
            return p
        else:
            p.status.append(PipelineStatus(state=State.READY,
                                           runner=runner,
                                           session=None,
                                           started_at=started_at,
                                           finished_at=None,
                                           task_id=None,
                                           task_name=str(run_pipeline)))

            # Run blocking database create in thread pool
            p = await run_in_threadpool(info.context["request"].app.backend.create, p)
            if unique_paths:
                p = generate_unique_paths(p, unique_paths)
                # Run blocking database update in thread pool
                p = await run_in_threadpool(info.context["request"].app.backend.update, p)

            result = run_pipeline.delay(
                id=str(p.id),
                name=serial["name"],
                parameters=serial["parameters"],
                data_catalog=serial["data_catalog"],
                runner=runner,
                slices=d.get("slices", None),
                only_missing=d.get("only_missing", False)
            )

            logger.info(
                f"user={PERMISSIONS_CLASS.get_user_info(info)['email']}, action=create_pipeline, id={p.id}, name={p.name}, state=READY, task_id={result.task_id}")
            return p

    @strawberry.mutation(description="Update a pipeline.", extensions=[PermissionExtension(permissions=[PERMISSIONS_CLASS(action="update_pipeline")]), PipelineInputExtension()])
    async def update_pipeline(self, id: str, pipeline: PipelineInput, info: Info, unique_paths: Optional[List[str]] = None) -> Pipeline:

        try:
            # Run blocking database read in thread pool
            p = await run_in_threadpool(info.context["request"].app.backend.read, id=id)
            if p is None:
                raise InvalidPipeline(
                    f"Pipeline {id} does not exist in the project.")
        except Exception as e:
            raise InvalidPipeline(f"Error retrieving pipeline {id}: {e}")

        pipeline_input_dict = jsonable_encoder(pipeline)

        # Update pipeline with new pipeline input
        p.parameters = pipeline_input_dict.get("parameters")
        p.data_catalog = pipeline_input_dict.get("data_catalog")
        p.tags = pipeline_input_dict.get("tags")
        p.parent = pipeline_input_dict.get("parent")
        runner = pipeline_input_dict.get(
            "runner") or info.context["request"].app.config["KEDRO_GRAPHQL_RUNNER"]

        # If PipelineInput is READY and pipeline is not already running
        if pipeline_input_dict.get("state", None) == "READY" and p.status[-1].state.value not in UNREADY_STATES.union(["READY"]):

            if (p.status[-1].state.value != "STAGED"):
                # Add new status object to pipeline because this is another run attempt
                p.status.append(PipelineStatus(state=State.READY,
                                               runner=runner,
                                               session=None,
                                               started_at=datetime.now(),
                                               finished_at=None,
                                               task_id=None,
                                               task_name=str(run_pipeline)))
            else:
                # Replace staged status with running status
                p.status[-1] = PipelineStatus(state=State.READY,
                                              runner=runner,
                                              session=None,
                                              started_at=datetime.now(),
                                              finished_at=None,
                                              task_id=None,
                                              task_name=str(run_pipeline))

            if unique_paths:
                p = generate_unique_paths(p, unique_paths)

            # Update pipeline in backend before running task
            # Run blocking database update in thread pool
            p = await run_in_threadpool(info.context["request"].app.backend.update, p)

            serial = p.encode(encoder="kedro")

            result = run_pipeline.delay(
                id=str(p.id),
                name=serial["name"],
                parameters=serial["parameters"],
                data_catalog=serial["data_catalog"],
                runner=runner,
                slices=pipeline_input_dict.get("slices", None),
                only_missing=pipeline_input_dict.get("only_missing", False)
            )

            logger.info(
                f"user={PERMISSIONS_CLASS.get_user_info(info)['email']}, action=run_pipeline, id={p.id}, name={p.name}, state=READY, task_id={result.task_id}")

        # If PipelineInput is STAGED and pipeline is not already running or staged
        elif pipeline_input_dict.get("state", None) == "STAGED" and p.status[-1].state.value not in UNREADY_STATES.union(["READY"]) and p.status[-1].state.value != "STAGED":
            p.status.append(PipelineStatus(state=State.STAGED,
                                           runner=runner,
                                           session=None,
                                           started_at=None,
                                           finished_at=None,
                                           task_id=None,
                                           task_name=None))
            logger.info(f'Staging pipeline {p.name}')
        if unique_paths:
            p = generate_unique_paths(p, unique_paths)
        # Run blocking database update in thread pool
        p = await run_in_threadpool(info.context["request"].app.backend.update, p)
        logger.info(
            f"user={PERMISSIONS_CLASS.get_user_info(info)['email']}, action=update_pipeline, id={p.id}, name={p.name}")

        return p

    @strawberry.mutation(description="Delete a pipeline.", extensions=[PermissionExtension(permissions=[PERMISSIONS_CLASS(action="delete_pipeline")]), PipelineExtension()])
    async def delete_pipeline(self, id: str, info: Info) -> Optional[Pipeline]:
        try:
            # Run blocking database read in thread pool
            p = await run_in_threadpool(info.context["request"].app.backend.read, id=id)
            if p is None:
                raise InvalidPipeline(
                    f"Pipeline {id} does not exist in the project.")
        except Exception as e:
            raise InvalidPipeline(f"Error retrieving pipeline {id}: {e}")

        # Run blocking database delete in thread pool
        await run_in_threadpool(info.context["request"].app.backend.delete, id=id)
        logger.info(f'Deleted {p.name} pipeline with id: ' + str(id))
        return p

    @strawberry.mutation(description="Create a dataset with a signed URL", extensions=[PermissionExtension(permissions=[PERMISSIONS_CLASS(action="create_dataset")])])
    async def create_datasets(self, id: str, info: Info, datasets: List[DataSetInput], expires_in_sec: int = CONFIG["KEDRO_GRAPHQL_SIGNED_URL_MAX_EXPIRES_IN_SEC"]) -> List[SignedUrl | SignedUrls | None]:
        """
        Get a signed URL for uploading a dataset.

        Args:
            id (str): The ID of the pipeline.
            info (Info): The GraphQL execution context.
            datasets (List[DataSetInput]): List of datasets for which to create signed URLs. In order to create specific partitions of a PartitionedDataset, pass a DataSetInput with the dataset name and list of partitions e.g. DataSetInput(name="dataset_name", partitions=["partition1", "partition2"]).
            expires_in_sec (int): The number of seconds the signed URL should be valid for.

        Returns:
            List[SignedUrl | SignedUrls | None]: A signed URL for uploading the dataset or None if not applicable.

        Raises:
            ValueError: If expires_in_sec is greater than max expires_in_sec
            DataSetConfigError: If the dataset configuration is invalid or cannot be parsed.
            TypeError: If the signed URL provider does not inherit from SignedUrlProvider.
        """
        if expires_in_sec > CONFIG["KEDRO_GRAPHQL_SIGNED_URL_MAX_EXPIRES_IN_SEC"]:
            raise ValueError(
                f"expires_in_sec cannot be greater than {CONFIG['KEDRO_GRAPHQL_SIGNED_URL_MAX_EXPIRES_IN_SEC']} seconds ({CONFIG['KEDRO_GRAPHQL_SIGNED_URL_MAX_EXPIRES_IN_SEC'] // 3600} hours)")
        urls = []
        # Run blocking database read in thread pool
        p = await run_in_threadpool(info.context["request"].app.backend.read, id=id)

        if p.status[-1].state.value != "STAGED":
            raise ValueError(
                f"Pipeline {p.name} with id {id} must be staged before creating datasets.")

        # create dict from pipeline data catalog
        catalog = {d.name: d for d in p.data_catalog}
        for dataset_input in datasets:
            dataset = catalog.get(dataset_input.name, None)
            if dataset is None:
                logger.warning(
                    f"Dataset '{dataset_input.name}' not found in the data catalog of pipeline_name={p.name} pipeline_id={p.id}. SignedURL set to None.")
                urls.append(None)
                continue
            else:
                module_path, class_name = CONFIG["KEDRO_GRAPHQL_SIGNED_URL_PROVIDER"].rsplit(
                    ".", 1)
                module = import_module(module_path)
                cls = getattr(module, class_name)

                if not issubclass(cls, SignedUrlProvider):
                    raise TypeError(
                        f"{class_name} must inherit from SignedUrlProvider")
                logger.info(
                    f"user={PERMISSIONS_CLASS.get_user_info(info)['email']}, action=create_dataset, expires_in_sec={expires_in_sec}")
                url = cls.create(info, dataset, expires_in_sec,
                                 dataset_input.partitions)
                urls.append(url)
        return urls


@strawberry.type
class Subscription:
    @strawberry.subscription(description="Subscribe to pipeline events.", extensions=[PermissionExtension(permissions=[PERMISSIONS_CLASS(action="subscribe_to_events")])])
    async def pipeline(self, id: str, info: Info, interval: float = 0.5) -> AsyncGenerator[PipelineEvent]:
        """Subscribe to pipeline events.
        """
        try:
            # Run blocking database read in thread pool
            p = await run_in_threadpool(info.context["request"].app.backend.read, id=id)
            if p is None:
                raise InvalidPipeline(
                    f"Pipeline {id} does not exist in the project.")
        except Exception as e:
            raise InvalidPipeline(f"Error retrieving pipeline {id}: {e}")

        while (not p.status[-1].task_id):
            # Wait for the task to be assigned a task_id
            await asyncio.sleep(0.1)
            # Run blocking database read in thread pool
            p = await run_in_threadpool(info.context["request"].app.backend.read, id=id)

        if p and p.status[-1].state.value not in READY_STATES:
            async for e in PipelineEventMonitor(app=info.context["request"].app.celery_app, task_id=p.status[-1].task_id).start(interval=interval):
                e["id"] = id
                yield PipelineEvent(**e)
        else:
            yield PipelineEvent(
                id=id,
                task_id=p.status[-1].task_id,
                timestamp=p.status[-1].finished_at,
                status=p.status[-1].state.value,
                result=p.status[-1].task_result,
                traceback=p.status[-1].task_traceback
            )

    @strawberry.subscription(description="Subscribe to pipeline logs.", extensions=[PermissionExtension(permissions=[PERMISSIONS_CLASS(action="subscribe_to_logs")])])
    async def pipeline_logs(self, id: str, info: Info) -> AsyncGenerator[PipelineLogMessage, None]:
        """Subscribe to pipeline logs."""
        try:
            # Run blocking database read in thread pool
            p = await run_in_threadpool(info.context["request"].app.backend.read, id=id)
            if p is None:
                raise InvalidPipeline(
                    f"Pipeline {id} does not exist in the project.")
        except Exception as e:
            raise InvalidPipeline(f"Error retrieving pipeline {id}: {e}")

        while (not p.status[-1].task_id):
            # Wait for the task to be assigned a task_id
            await asyncio.sleep(0.1)
            # Run blocking database read in thread pool
            p = await run_in_threadpool(info.context["request"].app.backend.read, id=id)

        if p:
            stream = await PipelineLogStream().create(task_id=p.status[-1].task_id, broker_url=info.context["request"].app.config["KEDRO_GRAPHQL_BROKER"])
            async for e in stream.consume():
                e["id"] = id
                yield PipelineLogMessage(**e)


def build_schema(
        type_plugins: dict[str, list],
        directives: Iterable[StrawberryDirective] = (),
        types: Iterable[Union[type, StrawberryType]] = (),
        extensions: Iterable[Union[type[SchemaExtension], SchemaExtension]] = (),
        execution_context_class: Optional[type[GraphQLExecutionContext]] = None,
        config: Optional[StrawberryConfig] = None,
        scalar_overrides: Optional[
            dict[object, Union[type, ScalarWrapper, ScalarDefinition]]
        ] = None,
        schema_directives: Iterable[object] = ()


):

    ComboQuery = merge_types("Query", tuple([Query] + type_plugins["query"]))
    ComboMutation = merge_types("Mutation", tuple(
        [Mutation] + type_plugins["mutation"]))
    ComboSubscription = merge_types("Subscription", tuple(
        [Subscription] + type_plugins["subscription"]))

    return strawberry.Schema(query=ComboQuery,
                             mutation=ComboMutation,
                             subscription=ComboSubscription,
                             directives=directives,
                             types=types,
                             extensions=extensions,
                             execution_context_class=execution_context_class,
                             config=config,
                             scalar_overrides=scalar_overrides,
                             schema_directives=schema_directives)



import os
import pathlib
from importlib import import_module

import click
import uvicorn
from kedro.framework.session import KedroSession
from kedro.framework.startup import bootstrap_project
from watchfiles import run_process

from .config import load_config, cli_config
from .logs.logger import logger


def init_app(app, config, session):
    # create app instance
    module, class_name = app.rsplit(".", 1)
    module = import_module(module)
    class_inst = getattr(module, class_name)
    return class_inst(kedro_session=session, config=config)


def start_app(app, config, conf_source, env, package_name, project_path):
    bootstrap_project(project_path)
    with KedroSession.create(project_path=project_path, env=env, conf_source=conf_source) as session:
        a = init_app(app, config, session)
        uvicorn.run(a, host="0.0.0.0", port=5000, log_level="info")


def start_worker(app, config, conf_source, env, package_name, project_path):

    bootstrap_project(project_path)
    with KedroSession.create(project_path=project_path, env=env, conf_source=conf_source) as session:
        a = init_app(app, config, session)
        worker = a.celery_app.Worker()
        worker.start()


@click.group(name="kedro-graphql")
def commands():
    pass


@commands.command()
@click.pass_obj
@click.option("--app", "-a", default=None, help="Application import path")
@click.option("--app-title", default=None, help="Title of the Kedro GraphQL application")
@click.option("--app-description", default=None, help="Description of the Kedro GraphQL application")
@click.option("--backend", default=None, help="The only supported value for this option is 'kedro_graphql.backends.mongodb.MongoBackend'")
@click.option("--broker", default=None, help="URI to broker e.g. 'redis://localhost'")
@click.option("--celery-result-backend", default=None, help="URI to backend for celery results e.g. 'redis://localhost'")
@click.option("--client-uri-graphql", default=None, help="URI for GraphQL API endpoint used by the GraphQL client")
@click.option("--client-uri-ws", default=None, help="URI for WebSocket endpoint used by the GraphQL client for subscriptions")
@click.option("--conf-source", default=None, help="Path of a directory where project configuration is stored.")
@click.option("--dataset-filepath-masks", default=None, help="List of masks to apply to Dataset filepaths before returning responses to client to hide true location of datasets (JSON string)")
@click.option("--dataset-filepath-allowed-roots", default=None, help="List of allowed root directories for Dataset filepaths (JSON string)")
@click.option("--deprecations-docs", default=None, help="URL to documentation about deprecated features")
@click.option("--env", "-e", default=None, help="Kedro configuration environment name. Defaults to `local`.")
@click.option("--events-config", default=None, help="Event configuration as JSON string")
@click.option("--imports", "-i", default=None, help="Additional import paths (comma-separated string or JSON array)")
@click.option("--local-file-provider-download-allowed-roots", default=None, help="Allowed root directories for downloads (comma-separated string or JSON array)")
@click.option("--local-file-provider-jwt-algorithm", default=None, help="Algorithm used for JWT signing (e.g., 'HS256')")
@click.option("--local-file-provider-jwt-secret-key", default=None, help="Secret key for signing JWT tokens for local file access")
@click.option("--local-file-provider-server-url", default=None, help="Base URL for the local file server")
@click.option("--local-file-provider-upload-allowed-roots", default=None, help="Allowed root directories for uploads (comma-separated string or JSON array)")
@click.option("--local-file-provider-upload-max-file-size-mb", default=None, type=int, help="Maximum allowed upload file size in megabytes")
@click.option("--log-path-prefix", default=None, help="Prefix of path to save logs")
@click.option("--log-tmp-dir", default=None, help="Temporary directory for logs")
@click.option("--mongo-db-collection", default=None, help="Name of the MongoDB collection to use")
@click.option("--mongo-db-name", default=None, help="Name to use for collection in mongo e.g. 'pipelines'")
@click.option("--mongo-uri", default=None, help="URI to mongodb e.g. 'mongodb://root:example@localhost:27017/'")
@click.option("--permissions", default=None, help="Python path to the permissions class used for authentication")
@click.option("--permissions-group-to-role-map", default=None, help="Mapping of external group names to roles as JSON string")
@click.option("--permissions-role-to-action-map", default=None, help="Mapping of roles to allowed actions as JSON string")
@click.option("--project-version", default=None, help="Version of the Kedro GraphQL project")
@click.option("--root-path", default=None, help="Root path for API endpoints (e.g., '/api/v1')")
@click.option("--runner", default=None, help="Execution mechanism to run pipelines e.g. 'kedro.runner.SequentialRunner'")
@click.option("--signed-url-max-expires-in-sec", default=None, type=int, help="Maximum allowed expiration time (in seconds) for presigned URLs")
@click.option("--signed-url-provider", default=None, help="Python path to the presigned URL provider class")
@click.option("--reload", "-r", is_flag=True, default=False, help="Enable auto-reload.")
@click.option("--reload-path", default=None, type=click.Path(exists=True, resolve_path=True, path_type=pathlib.Path), help="Path to watch for file changes, defaults to <project path>/src")
@click.option("--api-spec", default=None, type=click.Path(exists=True, resolve_path=True, path_type=pathlib.Path), help="Path to YAML API specification file")
@click.option("--ui", "-u", is_flag=True, default=False, help="Start a viz app.")
@click.option("--ui-spec", default="", help="UI YAML specification file")
@click.option("--worker", "-w", is_flag=True, default=False, help="Start a celery worker.")
def gql(metadata, app, app_title, app_description, backend, broker, celery_result_backend, client_uri_graphql, client_uri_ws, conf_source,
        dataset_filepath_masks, dataset_filepath_allowed_roots, deprecations_docs, env, events_config, imports,
        local_file_provider_download_allowed_roots,
        local_file_provider_jwt_algorithm, local_file_provider_jwt_secret_key, local_file_provider_server_url,
        local_file_provider_upload_allowed_roots, local_file_provider_upload_max_file_size_mb,
        log_path_prefix, log_tmp_dir, mongo_db_collection, mongo_db_name, mongo_uri, permissions,
        permissions_group_to_role_map, permissions_role_to_action_map, project_version, root_path, runner,
        signed_url_max_expires_in_sec, signed_url_provider,
        reload, reload_path, api_spec, ui, ui_spec, worker):
    """Commands for working with kedro-graphql."""

    # inject CLI options into config before calling load_config() to ensure all modules get same config
    if api_spec:
        os.environ["KEDRO_GRAPHQL_API_SPEC"] = str(api_spec)
    if app:
        cli_config["KEDRO_GRAPHQL_APP"] = app
    if app_title:
        cli_config["KEDRO_GRAPHQL_APP_TITLE"] = app_title
    if app_description:
        cli_config["KEDRO_GRAPHQL_APP_DESCRIPTION"] = app_description
    if backend:
        cli_config["KEDRO_GRAPHQL_BACKEND"] = backend
    if broker:
        cli_config["KEDRO_GRAPHQL_BROKER"] = broker
    if celery_result_backend:
        cli_config["KEDRO_GRAPHQL_CELERY_RESULT_BACKEND"] = celery_result_backend
    if client_uri_graphql:
        cli_config["KEDRO_GRAPHQL_CLIENT_URI_GRAPHQL"] = client_uri_graphql
    if client_uri_ws:
        cli_config["KEDRO_GRAPHQL_CLIENT_URI_WS"] = client_uri_ws
    if conf_source:
        cli_config["KEDRO_GRAPHQL_CONF_SOURCE"] = conf_source
    if dataset_filepath_masks:
        cli_config["KEDRO_GRAPHQL_DATASET_FILEPATH_MASKS"] = dataset_filepath_masks
    if dataset_filepath_allowed_roots:
        cli_config["KEDRO_GRAPHQL_DATASET_FILEPATH_ALLOWED_ROOTS"] = dataset_filepath_allowed_roots
    if deprecations_docs:
        cli_config["KEDRO_GRAPHQL_DEPRECATIONS_DOCS"] = deprecations_docs
    if env:
        cli_config["KEDRO_GRAPHQL_ENV"] = env
    if events_config:
        cli_config["KEDRO_GRAPHQL_EVENTS_CONFIG"] = events_config
    if imports:
        cli_config["KEDRO_GRAPHQL_IMPORTS"] = imports
    if local_file_provider_download_allowed_roots:
        cli_config["KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_DOWNLOAD_ALLOWED_ROOTS"] = local_file_provider_download_allowed_roots
    if local_file_provider_jwt_algorithm:
        cli_config["KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_JWT_ALGORITHM"] = local_file_provider_jwt_algorithm
    if local_file_provider_jwt_secret_key:
        cli_config["KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_JWT_SECRET_KEY"] = local_file_provider_jwt_secret_key
    if local_file_provider_server_url:
        cli_config["KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_SERVER_URL"] = local_file_provider_server_url
    if local_file_provider_upload_allowed_roots:
        cli_config["KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_UPLOAD_ALLOWED_ROOTS"] = local_file_provider_upload_allowed_roots
    if local_file_provider_upload_max_file_size_mb is not None:
        cli_config["KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_UPLOAD_MAX_FILE_SIZE_MB"] = local_file_provider_upload_max_file_size_mb
    if log_path_prefix:
        cli_config["KEDRO_GRAPHQL_LOG_PATH_PREFIX"] = log_path_prefix
    if log_tmp_dir:
        cli_config["KEDRO_GRAPHQL_LOG_TMP_DIR"] = log_tmp_dir
    if mongo_db_collection:
        cli_config["KEDRO_GRAPHQL_MONGO_DB_COLLECTION"] = mongo_db_collection
    if mongo_db_name:
        cli_config["KEDRO_GRAPHQL_MONGO_DB_NAME"] = mongo_db_name
    if mongo_uri:
        cli_config["KEDRO_GRAPHQL_MONGO_URI"] = mongo_uri
    if permissions:
        cli_config["KEDRO_GRAPHQL_PERMISSIONS"] = permissions
    if permissions_group_to_role_map:
        cli_config["KEDRO_GRAPHQL_PERMISSIONS_GROUP_TO_ROLE_MAP"] = permissions_group_to_role_map
    if permissions_role_to_action_map:
        cli_config["KEDRO_GRAPHQL_PERMISSIONS_ROLE_TO_ACTION_MAP"] = permissions_role_to_action_map
    if project_version:
        cli_config["KEDRO_GRAPHQL_PROJECT_VERSION"] = project_version
    if root_path:
        cli_config["KEDRO_GRAPHQL_ROOT_PATH"] = root_path
    if runner:
        cli_config["KEDRO_GRAPHQL_RUNNER"] = runner
    if signed_url_max_expires_in_sec is not None:
        cli_config["KEDRO_GRAPHQL_SIGNED_URL_MAX_EXPIRES_IN_SEC"] = signed_url_max_expires_in_sec
    if signed_url_provider:
        cli_config["KEDRO_GRAPHQL_SIGNED_URL_PROVIDER"] = signed_url_provider

    os.environ["KEDRO_GRAPHQL_PROJECT_VERSION"] = getattr(
        import_module(metadata.package_name), "__version__", None)
    os.environ["KEDRO_PROJECT_NAME"] = metadata.package_name

    config = load_config()
    logger.debug("configuration loaded by {s}".format(s=__name__))

    if not reload_path:
        reload_path = metadata.project_path.joinpath("src")

    if reload:
        logger.info("AUTO-RELOAD ACTIVATED, watching '" +
                    str(reload_path) + "' for changes")

    if ui:
        from .ui.app import start_ui
        if reload:
            run_process(str(reload_path), target=start_ui, kwargs={
                        "spec": ui_spec, "config": config})
        else:
            start_ui(spec=ui_spec, config=config)

    elif worker:
        if reload:
            run_process(str(reload_path), target=start_worker, args=(
                config["KEDRO_GRAPHQL_APP"], config, config["KEDRO_GRAPHQL_CONF_SOURCE"], config["KEDRO_GRAPHQL_ENV"], metadata.package_name, metadata.project_path))
        else:
            start_worker(config["KEDRO_GRAPHQL_APP"], config, config["KEDRO_GRAPHQL_CONF_SOURCE"], config["KEDRO_GRAPHQL_ENV"],
                         metadata.package_name, metadata.project_path)

    else:
        if reload:
            run_process(reload_path, target=start_app, args=(config["KEDRO_GRAPHQL_APP"], config, config["KEDRO_GRAPHQL_CONF_SOURCE"],
                        config["KEDRO_GRAPHQL_ENV"], metadata.package_name, metadata.project_path))
        else:
            start_app(config["KEDRO_GRAPHQL_APP"], config, config["KEDRO_GRAPHQL_CONF_SOURCE"], config["KEDRO_GRAPHQL_ENV"],
                      metadata.package_name, metadata.project_path)



from importlib import import_module


def init_backend(config):
    backend_module, backend_class = config["KEDRO_GRAPHQL_BACKEND"].rsplit(".", 1)
    backend_kwargs = {"uri": config.get("KEDRO_GRAPHQL_MONGO_URI"), "db": config.get(
        "KEDRO_GRAPHQL_MONGO_DB_NAME"), "collection": config.get("KEDRO_GRAPHQL_MONGO_DB_COLLECTION")}
    backend_module = import_module(backend_module)
    backend = getattr(backend_module, backend_class)
    return backend(**backend_kwargs)



# some code commented out because it has not been implemented
import panel as pn
import param
import pandas as pd
import json
from kedro_graphql.models import Pipeline

pn.extension('ace', 'jsoneditor')
pn.extension('tabulator', css_files=[
             "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.4.2/css/all.min.css"])


class PipelineDetail(pn.viewable.Viewer):
    pipeline = param.ClassSelector(class_=Pipeline)

    async def build_detail(self, raw):
        """Builds the detail view for the pipeline, showing its parameters, tags, data catalog, and status.

        TO DO:
            - Implement a more user-friendly way to view and interact with the pipeline's data catalog.
            - Support for "batch" view. View child pipelines or parent pipelines if applicable.

        Args:
            raw (pn.widgets.CheckButtonGroup): A widget to toggle between raw JSON view and structured view.
        Yields:
            pn.Row: A row containing the detail view of the pipeline.
        """
        if "json" in raw:
            yield pn.Row(
                pn.widgets.JSONEditor(
                    value=self.pipeline.encode(encoder="dict"),
                    mode="view",
                    width=600
                )
            )
        else:
            if self.pipeline.parameters:
                param_df = pd.DataFrame({
                    'name': [i.name for i in self.pipeline.parameters],
                    'value': [i.value for i in self.pipeline.parameters],
                }, index=[i for i in range(len(self.pipeline.parameters))])
            else:
                param_df = pd.DataFrame({
                    'name': [],
                    'value': [],
                }, index=[])

            param_widget = pn.widgets.Tabulator(
                param_df,
                disabled=True,
                theme='materialize',
                show_index=False
            )

            if self.pipeline.tags:
                tags_df = pd.DataFrame({
                    'key': [i.key for i in self.pipeline.tags],
                    'value': [i.value for i in self.pipeline.tags],
                }, index=[i for i in range(len(self.pipeline.tags))])
            else:
                tags_df = pd.DataFrame({'key': [], 'value': []})

            tags_widget = pn.widgets.Tabulator(
                tags_df,
                disabled=True,
                theme='materialize',
                show_index=False
            )

            if self.pipeline.data_catalog:
                ds_df = pd.DataFrame({
                    'name': [i.name for i in self.pipeline.data_catalog],
                    'type': [json.loads(i.config)["type"] for i in self.pipeline.data_catalog],
                }, index=[i for i in range(len(self.pipeline.data_catalog))])
            else:
                # backwards compatibility, will be deprecated
                ds_df = pd.DataFrame({
                    'name': [i.name for i in self.pipeline.inputs] + [i.name for i in self.pipeline.outputs],
                    'type': [i.type for i in self.pipeline.inputs] + [i.type for i in self.pipeline.outputs],
                }, index=[i for i in range(len(self.pipeline.inputs) + len(self.pipeline.outputs))])

            ds_widget = pn.widgets.Tabulator(
                ds_df,
                disabled=True,
                buttons={'Download': "<i class='fa fa-download'></i>"},
                theme='materialize',
                show_index=False
            )

            # in the future this will be a list of PipelineStatus objects
            # for now we grab each attribute from Pipeline object

            cols = [
                "state", "task_id", "task_name", "task_args", "task_kwargs",
                "task_request", "task_excpetion", "task_traceback", "task_einfo", "task_self.pipeline"
            ]
            values = self.pipeline.encode(encoder="dict")["status"]

            status_df = pd.DataFrame(
                values,
                columns=cols,
                index=[i for i in range(len(values))]
            )

            status_widget = pn.widgets.Tabulator(
                status_df,
                disabled=True,
                theme='materialize',
                layout='fit_columns',
                show_index=False
            )

            yield pn.Column(
                pn.Row(
                    pn.Card(
                        tags_widget,
                        title="Tags",
                        sizing_mode="stretch_width"
                    )
                ),
                pn.Row(
                    pn.Card(
                        param_widget,
                        title="Parameters",
                        sizing_mode="stretch_width"
                    )
                ),
                pn.Row(
                    pn.Card(
                        ds_widget,
                        title="Data Catalog",
                        sizing_mode="stretch_width"
                    )
                ),
                pn.Row(
                    pn.Card(
                        status_widget,
                        title="Runs",
                        sizing_mode="stretch_width"
                    )
                )
            )

    def __panel__(self):
        raw = pn.widgets.CheckButtonGroup(name='JSON', value=[], options=["json"])
        detail = pn.bind(self.build_detail, raw)
        return pn.Column(
            pn.Row(raw),
            pn.Row(detail),
            sizing_mode="stretch_width"
        )



"""
Tests for the configuration system.

This module tests configuration precedence, complex data type parsing,
and edge cases for the kedro-graphql configuration system.
"""
import json
import os
import yaml
from unittest.mock import patch, mock_open
import pytest

from kedro_graphql.config import load_config, env_var_parser, load_api_spec, defaults


class TestConfigurationPrecedence:
    """Test configuration precedence order."""

    def test_defaults_only(self):
        """Test that defaults work when no other config is provided."""
        with patch.dict(os.environ, {}, clear=True), \
                patch('kedro_graphql.config.dotenv_values', return_value={}), \
                patch('kedro_graphql.config.load_api_spec', return_value={}):

            config = load_config()
            assert config["KEDRO_GRAPHQL_APP_TITLE"] == "Kedro GraphQL API"
            assert config["KEDRO_GRAPHQL_IMPORTS"] == ["kedro_graphql.plugins.plugins"]

    def test_env_var_overrides_defaults(self):
        """Test that environment variables override defaults."""
        env_vars = {
            "KEDRO_GRAPHQL_APP_TITLE": "Custom Title from Env",
            "KEDRO_GRAPHQL_MONGO_URI": "mongodb://custom:27017/"
        }

        with patch.dict(os.environ, env_vars, clear=True), \
                patch('kedro_graphql.config.dotenv_values', return_value={}), \
                patch('kedro_graphql.config.load_api_spec', return_value={}):

            config = load_config()
            assert config["KEDRO_GRAPHQL_APP_TITLE"] == "Custom Title from Env"
            assert config["KEDRO_GRAPHQL_MONGO_URI"] == "mongodb://custom:27017/"
            # Defaults should still be there for non-overridden values
            assert config["KEDRO_GRAPHQL_BACKEND"] == defaults["KEDRO_GRAPHQL_BACKEND"]

    def test_dotenv_overrides_defaults(self):
        """Test that .env file values override defaults."""
        dotenv_values = {
            "KEDRO_GRAPHQL_APP_TITLE": "Title from dotenv",
            "KEDRO_GRAPHQL_BROKER": "redis://dotenv:6379"
        }

        with patch.dict(os.environ, {}, clear=True), \
                patch('kedro_graphql.config.dotenv_values', return_value=dotenv_values), \
                patch('kedro_graphql.config.load_api_spec', return_value={}):

            config = load_config()
            assert config["KEDRO_GRAPHQL_APP_TITLE"] == "Title from dotenv"
            assert config["KEDRO_GRAPHQL_BROKER"] == "redis://dotenv:6379"

    def test_env_var_overrides_dotenv(self):
        """Test that environment variables override .env file values."""
        dotenv_values = {
            "KEDRO_GRAPHQL_APP_TITLE": "Title from dotenv",
            "KEDRO_GRAPHQL_BROKER": "redis://dotenv:6379"
        }
        env_vars = {
            "KEDRO_GRAPHQL_APP_TITLE": "Title from env var"
        }

        with patch.dict(os.environ, env_vars, clear=True), \
                patch('kedro_graphql.config.dotenv_values', return_value=dotenv_values), \
                patch('kedro_graphql.config.load_api_spec', return_value={}):

            config = load_config()
            # env var wins
            assert config["KEDRO_GRAPHQL_APP_TITLE"] == "Title from env var"
            # dotenv for non-conflicting
            assert config["KEDRO_GRAPHQL_BROKER"] == "redis://dotenv:6379"

    def test_cli_config_overrides_env_var(self):
        """Test that CLI config overrides environment variables."""
        env_vars = {
            "KEDRO_GRAPHQL_APP_TITLE": "Title from env var",
            "KEDRO_GRAPHQL_MONGO_URI": "mongodb://env:27017/"
        }
        cli_config = {
            "KEDRO_GRAPHQL_APP_TITLE": "Title from CLI"
        }

        with patch.dict(os.environ, env_vars, clear=True), \
                patch('kedro_graphql.config.dotenv_values', return_value={}), \
                patch('kedro_graphql.config.load_api_spec', return_value={}):

            config = load_config(cli_config=cli_config)
            assert config["KEDRO_GRAPHQL_APP_TITLE"] == "Title from CLI"  # CLI wins
            # env for non-conflicting
            assert config["KEDRO_GRAPHQL_MONGO_URI"] == "mongodb://env:27017/"

    def test_yaml_spec_overrides_cli_config(self):
        """Test that YAML spec has highest precedence."""
        env_vars = {
            "KEDRO_GRAPHQL_APP_TITLE": "Title from env var"
        }
        cli_config = {
            "KEDRO_GRAPHQL_APP_TITLE": "Title from CLI",
            "KEDRO_GRAPHQL_BROKER": "redis://cli:6379"
        }
        yaml_config = {
            "KEDRO_GRAPHQL_APP_TITLE": "Title from YAML"
        }

        with patch.dict(os.environ, env_vars, clear=True), \
                patch('kedro_graphql.config.dotenv_values', return_value={}), \
                patch('kedro_graphql.config.load_api_spec', return_value=yaml_config):

            config = load_config(cli_config=cli_config)
            assert config["KEDRO_GRAPHQL_APP_TITLE"] == "Title from YAML"  # YAML wins
            # CLI for non-conflicting
            assert config["KEDRO_GRAPHQL_BROKER"] == "redis://cli:6379"

    def test_full_precedence_chain(self):
        """Test complete precedence chain with all sources."""
        dotenv_values = {
            "KEDRO_GRAPHQL_APP_TITLE": "Title from dotenv",
            "KEDRO_GRAPHQL_BROKER": "redis://dotenv:6379",
            "KEDRO_GRAPHQL_BACKEND": "backend.from.dotenv"
        }
        env_vars = {
            "KEDRO_GRAPHQL_APP_TITLE": "Title from env",
            "KEDRO_GRAPHQL_BROKER": "redis://env:6379"
        }
        cli_config = {
            "KEDRO_GRAPHQL_APP_TITLE": "Title from CLI"
        }
        yaml_config = {
            "KEDRO_GRAPHQL_MONGO_URI": "mongodb://yaml:27017/"
        }

        with patch.dict(os.environ, env_vars, clear=True), \
                patch('kedro_graphql.config.dotenv_values', return_value=dotenv_values), \
                patch('kedro_graphql.config.load_api_spec', return_value=yaml_config):

            config = load_config(cli_config=cli_config)

            # YAML spec (highest precedence)
            assert config["KEDRO_GRAPHQL_MONGO_URI"] == "mongodb://yaml:27017/"

            # CLI config (second highest)
            assert config["KEDRO_GRAPHQL_APP_TITLE"] == "Title from CLI"

            # Environment variable (third)
            assert config["KEDRO_GRAPHQL_BROKER"] == "redis://env:6379"

            # .env file (fourth)
            assert config["KEDRO_GRAPHQL_BACKEND"] == "backend.from.dotenv"


class TestComplexDataTypeParsing:
    """Test parsing of complex data types (lists, dictionaries)."""

    def test_list_field_comma_separated_string(self):
        """Test parsing comma-separated strings for list fields."""
        config = {
            "KEDRO_GRAPHQL_IMPORTS": "module1.plugin,module2.plugin,module3.plugin",
            "KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_DOWNLOAD_ALLOWED_ROOTS": "./data,/var,/tmp"
        }

        parsed = env_var_parser(config)

        assert parsed["KEDRO_GRAPHQL_IMPORTS"] == [
            "module1.plugin", "module2.plugin", "module3.plugin"]
        assert parsed["KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_DOWNLOAD_ALLOWED_ROOTS"] == [
            "./data", "/var", "/tmp"]

    def test_list_field_json_array(self):
        """Test parsing JSON arrays for list fields."""
        config = {
            "KEDRO_GRAPHQL_IMPORTS": '["module1.plugin", "module2.plugin"]',
            "KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_UPLOAD_ALLOWED_ROOTS": '["./data", "./uploads"]'
        }

        parsed = env_var_parser(config)

        assert parsed["KEDRO_GRAPHQL_IMPORTS"] == ["module1.plugin", "module2.plugin"]
        assert parsed["KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_UPLOAD_ALLOWED_ROOTS"] == [
            "./data", "./uploads"]

    def test_list_field_single_value(self):
        """Test parsing single values that should become single-item lists."""
        config = {
            "KEDRO_GRAPHQL_IMPORTS": "single.module",
            "KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_UPLOAD_ALLOWED_ROOTS": "./uploads"
        }

        parsed = env_var_parser(config)

        assert parsed["KEDRO_GRAPHQL_IMPORTS"] == ["single.module"]
        assert parsed["KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_UPLOAD_ALLOWED_ROOTS"] == [
            "./uploads"]

    def test_list_field_empty_string(self):
        """Test parsing empty strings for list fields."""
        config = {
            "KEDRO_GRAPHQL_IMPORTS": "",
            "KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_DOWNLOAD_ALLOWED_ROOTS": "   "  # whitespace only
        }

        parsed = env_var_parser(config)

        assert parsed["KEDRO_GRAPHQL_IMPORTS"] == []
        assert parsed["KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_DOWNLOAD_ALLOWED_ROOTS"] == []

    def test_list_field_mixed_whitespace(self):
        """Test parsing comma-separated strings with mixed whitespace."""
        config = {
            "KEDRO_GRAPHQL_IMPORTS": " module1 , module2,  module3  ,, module4 ",
            "KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_DOWNLOAD_ALLOWED_ROOTS": "./data, /var , /tmp,"
        }

        parsed = env_var_parser(config)

        # Empty elements should be filtered out
        assert parsed["KEDRO_GRAPHQL_IMPORTS"] == [
            "module1", "module2", "module3", "module4"]
        assert parsed["KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_DOWNLOAD_ALLOWED_ROOTS"] == [
            "./data", "/var", "/tmp"]

    def test_json_field_dictionary(self):
        """Test parsing JSON dictionaries for complex fields."""
        config = {
            "KEDRO_GRAPHQL_EVENTS_CONFIG": '{"event1": {"source": "app", "type": "test"}}',
            "KEDRO_GRAPHQL_PERMISSIONS_ROLE_TO_ACTION_MAP": '{"admin": ["create", "read"], "viewer": ["read"]}'
        }

        parsed = env_var_parser(config)

        expected_events = {"event1": {"source": "app", "type": "test"}}
        expected_permissions = {"admin": ["create", "read"], "viewer": ["read"]}

        assert parsed["KEDRO_GRAPHQL_EVENTS_CONFIG"] == expected_events
        assert parsed["KEDRO_GRAPHQL_PERMISSIONS_ROLE_TO_ACTION_MAP"] == expected_permissions

    def test_json_field_invalid_json(self):
        """Test handling of invalid JSON in JSON fields."""
        config = {
            "KEDRO_GRAPHQL_EVENTS_CONFIG": '{"invalid": json}',  # Invalid JSON
            "KEDRO_GRAPHQL_PERMISSIONS_GROUP_TO_ROLE_MAP": 'not json at all'
        }

        with patch('kedro_graphql.config.logger') as mock_logger:
            parsed = env_var_parser(config)

            # Should remain as original string when JSON parsing fails
            assert parsed["KEDRO_GRAPHQL_EVENTS_CONFIG"] == '{"invalid": json}'
            assert parsed["KEDRO_GRAPHQL_PERMISSIONS_GROUP_TO_ROLE_MAP"] == 'not json at all'

            # Should log warnings
            assert mock_logger.warning.call_count == 2

    def test_list_field_complex_json_array(self):
        """Test parsing complex JSON arrays with nested objects."""
        config = {
            "KEDRO_GRAPHQL_IMPORTS": '[{"module": "plugin1", "config": {"enabled": true}}, "simple.module"]'
        }

        parsed = env_var_parser(config)

        expected = [{"module": "plugin1", "config": {"enabled": True}}, "simple.module"]
        assert parsed["KEDRO_GRAPHQL_IMPORTS"] == expected

    def test_non_string_values_remain_unchanged(self):
        """Test that non-string values in config remain unchanged."""
        config = {
            "KEDRO_GRAPHQL_IMPORTS": ["already", "a", "list"],
            "KEDRO_GRAPHQL_EVENTS_CONFIG": {"already": "a", "dict": True},
            "KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_UPLOAD_MAX_FILE_SIZE_MB": 10,  # integer
        }

        parsed = env_var_parser(config)

        # Should remain as-is
        assert parsed["KEDRO_GRAPHQL_IMPORTS"] == ["already", "a", "list"]
        assert parsed["KEDRO_GRAPHQL_EVENTS_CONFIG"] == {"already": "a", "dict": True}
        assert parsed["KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_UPLOAD_MAX_FILE_SIZE_MB"] == 10


class TestYAMLSpecLoading:
    """Test YAML specification file loading."""

    def test_no_api_spec_env_var(self):
        """Test behavior when no API spec environment variable is set."""
        with patch.dict(os.environ, {}, clear=True):
            result = load_api_spec()
            assert result == {}

    def test_api_spec_file_not_found(self):
        """Test behavior when API spec file doesn't exist."""
        with patch.dict(os.environ, {"KEDRO_GRAPHQL_API_SPEC": "/nonexistent/file.yaml"}):
            with pytest.raises(FileNotFoundError):
                load_api_spec()

    def test_api_spec_invalid_yaml(self):
        """Test behavior with invalid YAML content."""
        invalid_yaml = "invalid: yaml: content: [unclosed"

        with patch.dict(os.environ, {"KEDRO_GRAPHQL_API_SPEC": "test.yaml"}), \
                patch("builtins.open", mock_open(read_data=invalid_yaml)), \
                patch('kedro_graphql.config.logger') as mock_logger:

            # Should handle gracefully and return empty dict
            result = load_api_spec()
            assert result == {}

            # Should log the error
            mock_logger.error.assert_called_once()

    def test_api_spec_valid_yaml(self):
        """Test successful YAML spec loading."""
        yaml_content = """
config:
  app_title: "Custom App Title"
  mongo_uri: "mongodb://yaml:27017/"
  imports:
    - "custom.plugin1"
    - "custom.plugin2"
  events_config:
    event1:
      source: "yaml.source"
      type: "yaml.event"
"""

        with patch.dict(os.environ, {"KEDRO_GRAPHQL_API_SPEC": "test.yaml"}), \
                patch("builtins.open", mock_open(read_data=yaml_content)), \
                patch('kedro_graphql.config.import_module') as mock_import:

            result = load_api_spec()

            expected = {
                "KEDRO_GRAPHQL_APP_TITLE": "Custom App Title",
                "KEDRO_GRAPHQL_MONGO_URI": "mongodb://yaml:27017/",
                "KEDRO_GRAPHQL_IMPORTS": ["custom.plugin1", "custom.plugin2"],
                "KEDRO_GRAPHQL_EVENTS_CONFIG": {
                    "event1": {"source": "yaml.source", "type": "yaml.event"}
                }
            }

            assert result == expected

            # Should import the specified modules
            mock_import.assert_any_call("custom.plugin1")
            mock_import.assert_any_call("custom.plugin2")

    def test_api_spec_imports_comma_separated_string(self):
        """Test YAML spec with imports as comma-separated string."""
        yaml_content = """
config:
  imports: "plugin1,plugin2,plugin3"
  app_title: "Test App"
"""

        with patch.dict(os.environ, {"KEDRO_GRAPHQL_API_SPEC": "test.yaml"}), \
                patch("builtins.open", mock_open(read_data=yaml_content)), \
                patch('kedro_graphql.config.import_module') as mock_import:

            result = load_api_spec()

            assert result["KEDRO_GRAPHQL_IMPORTS"] == "plugin1,plugin2,plugin3"

            # Should import each module
            mock_import.assert_any_call("plugin1")
            mock_import.assert_any_call("plugin2")
            mock_import.assert_any_call("plugin3")


class TestEdgeCases:
    """Test edge cases and error conditions."""

    def test_empty_cli_config(self):
        """Test that empty cli_config works properly."""
        with patch.dict(os.environ, {}, clear=True), \
                patch('kedro_graphql.config.dotenv_values', return_value={}), \
                patch('kedro_graphql.config.load_api_spec', return_value={}):

            config = load_config(cli_config={})
            assert config["KEDRO_GRAPHQL_APP_TITLE"] == defaults["KEDRO_GRAPHQL_APP_TITLE"]



import asyncio
from typing import AsyncGenerator

import strawberry

from kedro_graphql.decorators import (
    gql_mutation,
    gql_query,
    gql_subscription,
)


@gql_query()
@strawberry.type
class ExampleQueryTypePlugin():
    @strawberry.field
    def hello_world(self) -> str:
        return "Hello World"


@gql_mutation()
@strawberry.type
class ExampleMutationTypePlugin():
    @strawberry.mutation
    def hello_world(self, message: str = "World") -> str:
        return "Hello " + message


@gql_subscription()
@strawberry.type
class ExampleSubscriptionTypePlugin():
    @strawberry.subscription
    async def hello_world(self, message: str = "World", target: int = 11) -> AsyncGenerator[str, None]:
        for i in range(target):
            yield str(i) + " Hello " + message
            await asyncio.sleep(0.5)



from fastapi.middleware.cors import CORSMiddleware

from kedro_graphql import KedroGraphQL


class MyApp(KedroGraphQL):

    def __init__(self):
        super(MyApp, self).__init__()

        origins = [
            "http://localhost",
            "http://localhost:8080",
        ]

        self.add_middleware(
            CORSMiddleware,
            allow_origins=origins,
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
        )
        print("added CORSMiddleware")



import asyncio
import json
import param
import panel as pn
import pandas as pd
from kedro_graphql.models import Pipeline, DataSetInput, SignedUrl, SignedUrls

pn.extension(notifications=True)


class DataCatalogExplorer(pn.viewable.Viewer):
    """
    A Panel component that displays the data catalog of a Kedro pipeline, allowing users to filter, view,
    and download datasets using their pre-signed URL implementation of choice.

    Attributes:
        pipeline (kedro_graphql.models.Pipeline): The Kedro pipeline object
        spec (dict): Kedro GraphQL UI specification
        dataset_map (dict): A mapping of Kedro dataset types to their corresponding rendering Panel components
    """

    pipeline = param.ClassSelector(class_=Pipeline, doc="The Kedro pipeline object")
    spec = param.Dict(default={}, doc="Kedro GraphQL UI specification")
    dataset_map = param.Dict(
        default={}, doc="A mapping of Kedro dataset types to their corresponding rendering Panel components")

    def __init__(self, **params):
        super().__init__(**params)

        # this is used to inject JS dynamically
        self.js_download = pn.pane.HTML("", width=0, height=0)
        self.partitions_widget = pn.widgets.Tabulator(
            pd.DataFrame({"Name": [], "Type": [], "Filepath": [], "Signed URL": []}),
            disabled=True,
            buttons={
                'Download': "<i class='fa fa-download'></i>",
                "Popout": "<i class='fa fa-external-link'></i>"
            },
            theme='materialize',
            selectable="checkbox",
            show_index=False,
            hidden_columns=["Signed URL", "Partitioned Dataset Name"]
        )
        self.partitions_download_button = pn.widgets.Button(
            name='Download', button_type='primary')
        self.partitions_download_button.on_click(lambda event: asyncio.create_task(
            self.on_partitions_download_click(event, self.partitions_widget.value)))
        self.partitions_modal = pn.Modal(self.partitions_download_button, self.partitions_widget, name="Partitions",
                                         margin=20, background_close=True, show_close_button=True)

        if not self.pipeline or not self.spec:
            self._content = pn.Column(
                pn.pane.Markdown(
                    "**Missing required parameters: `pipeline`, or `spec`.**"),
                sizing_mode="stretch_width"
            )
        else:
            # Show loading spinner until the dashboard is built
            self._content = pn.Column(
                pn.indicators.LoadingSpinner(value=True, width=50, height=50),
                pn.pane.Markdown("Fetching Data..."),
                sizing_mode='stretch_width'
            )

            # Ensures build after panel server is fully loaded to avoid race conditions
            pn.state.onload(lambda: asyncio.create_task(
                asyncio.to_thread(self.build_component)))

    def build_component(self):
        df_data = self._build_dataframe()
        ds_widget = self._build_widget(df_data)
        self.partitions_widget.on_click(lambda event: asyncio.create_task(
            self.on_partition_row_click(event, self.partitions_widget.value)))
        self._content[:] = pn.Column(
            pn.Card(ds_widget, title="Data Catalog"),
            self.js_download,
            self.partitions_modal,
            sizing_mode="stretch_width"
        )

    def _build_dataframe(self):
        base_df = pd.DataFrame({
            'Name': [ds.name for ds in self.pipeline.data_catalog],
            'Type': [ds.parse_config()["type"] for ds in self.pipeline.data_catalog],
            'Filepath': [ds.parse_config().get("filepath", ds.parse_config().get("path", "")) for ds in self.pipeline.data_catalog],
        })

        tags_list = []
        for idx, ds in enumerate(self.pipeline.data_catalog):
            tag_data = {}
            if ds.tags:
                for tag in ds.tags:
                    tag_data[f"Tag:{tag.key}"] = tag.value
            tags_list.append(pd.Series(tag_data, name=idx))

        tags_df = pd.DataFrame(tags_list).fillna("")
        return pd.concat([base_df, tags_df], axis=1)

    def _build_widget(self, ds_df):
        filters = {
            "Name": {"type": "input", "placeholder": "Filter by name", "func": "like"},
            "Type": {"type": "list", "placeholder": "Select dataset type(s)",
                     "func": "in", "valuesLookup": True, "multiselect": True},
            "Filepath": {"type": "input", "placeholder": "Filter by filepath", "func": "like"}
        }

        for tag in ds_df.columns[3:]:
            filters[tag] = {
                "type": "list",
                "placeholder": "Select tag(s)",
                "func": "in",
                "valuesLookup": True,
                "multiselect": True
            }

        ds_widget = pn.widgets.Tabulator(
            ds_df,
            disabled=True,
            buttons={
                'Download': "<i class='fa fa-download'></i>",
                "Popout": "<i class='fa fa-external-link'></i>"
            },
            theme='materialize',
            selectable=False,
            show_index=False,
            header_filters=filters,
        )
        ds_widget.on_click(lambda event: asyncio.create_task(
            self.on_row_click(event, ds_df)))
        return ds_widget

    def trigger_popout(self, url):
        """
        Trigger a popout window for the specified URL. Default browser rendering is used.
        """

        js_code = f"""
        <script>
        (function() {{
            const a = document.createElement('a');
            a.href = "{url}";
            a.target = "_blank";
            document.body.appendChild(a);
            a.click();
            document.body.removeChild(a);
        }})();
        </script>
        """

        self.js_download.object = js_code
        self.js_download.object = None

    def trigger_download(self, url, filepath):
        """
        Trigger a browser download for the specified URL. Name is derived from the filepath.
        """

        js_code = f"""
        <script>
        (async function() {{
            const response = await fetch("{url}");
            const blob = await response.blob();
            const url = window.URL.createObjectURL(blob);
            const link = document.createElement('a');
            link.href = url;
            link.setAttribute('download', "{filepath.split("/")[-1]}");
            document.body.appendChild(link);
            link.click();
            document.body.removeChild(link);
            window.URL.revokeObjectURL(url);
        }})();
        </script>
        """

        self.js_download.object = js_code
        self.js_download.object = None

    def open_dataset_viewer(self, page, presigned_url, ds_name, ds_type):
        """
        Open custom dataset viewer in a new browser tab provided from the dataset_map.
        """

        js_code = f"""
        <script>
        (function() {{
            const presigned_url = "{presigned_url}";
            const page = "{page}";
            const ds_name = "{ds_name}";
            const ds_type = "{ds_type}";
            const viewerUrl = `${{window.location.origin}}/?page=${{page}}&presigned_url=${{encodeURIComponent(presigned_url)}}&ds_name=${{encodeURIComponent(ds_name)}}&ds_type=${{encodeURIComponent(ds_type)}}`;
            window.open(viewerUrl, '_blank');
        }})();
        </script>
        """

        self.js_download.object = js_code
        self.js_download.object = None

    async def on_partitions_download_click(self, event, df):
        """
        Handle download button click events in the partitions table.
        """
        selected_rows = self.partitions_widget.selection
        for row in selected_rows:
            dataset_name = df.iloc[row]['Name']
            dataset_filepath = df.iloc[row]['Filepath']
            dataset_type = df.iloc[row]['Type']
            presigned_url = df.iloc[row]['Signed URL']
            self.trigger_download(
                presigned_url, dataset_filepath)

    async def on_partition_row_click(self, event, df):
        """
        Handle row click events in the partitions table.
        """
        row = event.row
        dataset_name = df.iloc[row]['Name']
        dataset_filepath = df.iloc[row]['Filepath']
        dataset_type = df.iloc[row]['Type']
        presigned_url = df.iloc[row]['Signed URL']
        partitioned_dataset_name = df.iloc[row]["Partitioned Dataset Name"]

        for ds in self.pipeline.data_catalog:
            if ds.name == partitioned_dataset_name:
                dataset = ds

        if event.column != 'Download' and event.column != 'Popout':
            return

        elif event.column == 'Download':
            self.trigger_download(
                presigned_url, dataset_filepath)

        elif event.column == 'Popout':
            # Check dataset_map for custom viewer
            for dataset_type, panel_page in self.dataset_map.items():
                if dataset.parse_config()["type"] == dataset_type:
                    self.open_dataset_viewer(
                        panel_page, presigned_url, dataset_name, dataset_type)
                    return
            # Default to raw data viewer
            self.trigger_popout(presigned_url)

    def open_partition_viewer(self, dataset, presigned_urls):
        ds_config = dataset.parse_config()
        self.partitions_widget.value = pd.DataFrame(
            {"Name": [url.file for url in presigned_urls.urls],
             "Type": [ds_config["dataset"]["type"] for url in presigned_urls.urls],
             "Signed URL": [u.url for u in presigned_urls.urls],
             "Partitioned Dataset Name": [dataset.name for u in presigned_urls.urls],
             "Filepath": ["/".join([ds_config.get("path", ""), url.file]) for url in presigned_urls.urls],

             })

        self.partitions_modal.show()

    async def on_row_click(self, event, ds_df):
        """
        Handle row click events in the dataset table and triggers appropriate actions for 'Download' or 'Popout' column clicks.
        """
        try:
            if event.column != 'Download' and event.column != 'Popout':
                return

            row = event.row
            dataset_name = ds_df.iloc[row]['Name']
            dataset_filepath = ds_df.iloc[row]['Filepath']
            dataset_type = ds_df.iloc[row]['Type']

            for ds in self.pipeline.data_catalog:
                if ds.name == dataset_name:

                    if dataset_type == "partitions.IncrementalDataset":
                        pn.state.notifications.warning(
                            f"Actions for IncrementalDatasets are not supported yet: {dataset_name}", duration=10000)
                        return

                    else:
                        presigned_urls = await self.spec["config"]["client"].read_datasets(
                            id=self.pipeline.id, datasets=[DataSetInput(name=ds.name)], expires_in_sec=3600)

                        presigned_url = presigned_urls[0]

                        if event.column == 'Download':
                            if dataset_type == "partitions.PartitionedDataset":
                                self.open_partition_viewer(ds, presigned_url)
                            elif isinstance(presigned_url, SignedUrl):
                                self.trigger_download(
                                    presigned_url.url, dataset_filepath)
                            else:
                                pn.state.notifications.warning(
                                    f"Download not available for {dataset_name}", duration=10000)
                            return

                        elif event.column == 'Popout':
                            if dataset_type == "partitions.PartitionedDataset":
                                self.open_partition_viewer(ds, presigned_url)
                            elif presigned_url:
                                # Check dataset_map for custom viewer
                                for dataset_type, panel_page in self.dataset_map.items():
                                    if ds.parse_config()["type"] == dataset_type:
                                        self.open_dataset_viewer(
                                            panel_page, presigned_url.url, dataset_name, dataset_type)
                                        return
                                # Default to raw data viewer
                                self.trigger_popout(presigned_url.url)
                            else:
                                pn.state.notifications.warning(
                                    f"No signed URL available for {dataset_name}", duration=10000)
                            return
        except Exception as e:
            pn.state.notifications.error(
                f"Error processing action for dataset {dataset_name}: {str(e)}", duration=10000)

    def __panel__(self):
        return self._content



import pytest

from requests import Request, Session
from requests.adapters import HTTPAdapter
from urllib3.util import Retry
from cloudevents.http import CloudEvent
from cloudevents.conversion import to_structured


@pytest.fixture
def mock_payload():
    query = """
    query TestQuery($limit: Int!) {
      pipelineTemplates(limit: $limit) {
        pageMeta {
          nextCursor
        }
        pipelineTemplates {
          name
          describe
          inputs {
            name
          }
          nodes {
            name
            inputs
            outputs
            tags
          }
          outputs {
            name
          }
          parameters {
            name
            value
          }
        }
      }
    }
    """
    variables = {"limit": 5}
    return {"query": query, "variables": variables}


class TestIsAuthenticatedXForwarded:

    def event_data(self):
        attributes = {"id": "1234",
                      "type": "com.example.event",
                      "source": "example.com"}
        data = {"key": "value"}
        return CloudEvent(attributes, data)

    def event_post(self, port=5000, headers={}):
        retry = Retry(
            total=5,
            backoff_factor=0.5,
            status_forcelist=[429, 500, 502, 503, 504],
        )

        adapter = HTTPAdapter(max_retries=retry)
        s = Session()
        s.mount('http://', adapter)

        event_headers, body = to_structured(self.event_data())
        headers.update(event_headers)

        url = "http://localhost:"+str(port)+"/event/"
        req = Request('POST', url, data=body, headers=headers)
        req_prepped = req.prepare()
        return s.send(req_prepped)

    @pytest.mark.asyncio
    async def test_rest_with_x_forwarded_groups_header(self,
                                                       mock_server_5001):
        headers = {
            "X-Forwarded-Groups": "test_group",
            "X-Forwarded-Email": "admin@example.com"}

        resp = self.event_post(port=5001, headers=headers)
        assert resp.status_code == 200

    @pytest.mark.asyncio
    async def test_rest_without_x_forwarded_groups_header(self,
                                                          mock_server_5001):
        headers = {
            "X-Forwarded-Email": "admin@example.com"}

        resp = self.event_post(port=5001, headers=headers)
        assert resp.status_code == 403

    @pytest.mark.asyncio
    async def test_graphql_with_x_forwarded_groups_header(self,
                                                          mock_payload,
                                                          mock_server_5001):

        retry = Retry(
            total=5,
            backoff_factor=0.5,
            status_forcelist=[429, 500, 502, 503, 504],
        )

        adapter = HTTPAdapter(max_retries=retry)
        s = Session()
        s.mount('http://', adapter)
        headers = {
            "X-Forwarded-Groups": "test_group",
            "X-Forwarded-Email": "admin@example.com"}
        url = "http://localhost:5001/graphql"
        req = Request('POST', url, json=mock_payload, headers=headers)
        req_prepped = req.prepare()
        # print("REQUEST:", req_prepped.headers, req_prepped.body)
        resp = s.send(req_prepped)
        # print("RESPONSE:", resp)
        assert resp.status_code == 200
        resp = resp.json()
        assert resp.keys() == {"data"}
        assert resp["data"].keys() == {"pipelineTemplates"}

    @pytest.mark.asyncio
    async def test_graphql_without_x_forwarded_groups_header(self,
                                                             mock_payload,
                                                             mock_server_5001):

        retry = Retry(
            total=5,
            backoff_factor=0.5,
            status_forcelist=[429, 500, 502, 503, 504],
        )

        adapter = HTTPAdapter(max_retries=retry)
        s = Session()
        s.mount('http://', adapter)

        url = "http://localhost:5001/graphql"
        req = Request('POST', url, json=mock_payload, headers={})
        req_prepped = req.prepare()
        # print("REQUEST:", req_prepped.headers, req_prepped.body)
        resp = s.send(req_prepped)
        # print("RESPONSE:", resp)
        assert resp.status_code == 200
        resp = resp.json()
        assert resp.keys() == {"data", "errors"}
        assert resp["errors"][0]["message"] == "User is not authenticated"

    @pytest.mark.asyncio
    async def test_graphql_with_wrong_x_forwarded_groups_header(self,
                                                                mock_payload,
                                                                mock_server_5001):

        retry = Retry(
            total=5,
            backoff_factor=0.5,
            status_forcelist=[429, 500, 502, 503, 504],
        )

        adapter = HTTPAdapter(max_retries=retry)
        s = Session()
        s.mount('http://', adapter)

        headers = {
            "X-Forwarded-Groups": "wrong_group",
            "X-Forwarded-Email": "admin@example.com"}
        url = "http://localhost:5001/graphql"
        req = Request('POST', url, json=mock_payload, headers=headers)

        req_prepped = req.prepare()
        # print("REQUEST:", req_prepped.headers, req_prepped.body)
        resp = s.send(req_prepped)
        # print("RESPONSE:", resp)
        assert resp.status_code == 200
        resp = resp.json()
        assert resp.keys() == {"data", "errors"}
        assert resp["errors"][0]["message"] == "User is not authenticated"

    @pytest.mark.asyncio
    async def test_graphql_with_x_forwarded_email(self,
                                                  mock_payload,
                                                  mock_server_5002):

        retry = Retry(
            total=5,
            backoff_factor=0.5,
            status_forcelist=[429, 500, 502, 503, 504],
        )

        adapter = HTTPAdapter(max_retries=retry)
        s = Session()
        s.mount('http://', adapter)
        headers = {
            "X-Forwarded-Email": "admin@example.com"}
        url = "http://localhost:5002/graphql"
        req = Request('POST', url, json=mock_payload, headers=headers)
        req_prepped = req.prepare()
        # print("REQUEST:", req_prepped.headers, req_prepped.body)
        resp = s.send(req_prepped)
        # print("RESPONSE:", resp)
        assert resp.status_code == 200
        resp = resp.json()
        assert resp.keys() == {"data"}
        assert resp["data"].keys() == {"pipelineTemplates"}

    @pytest.mark.asyncio
    async def test_graphql_without_x_forwarded_email(self,
                                                     mock_payload,
                                                     mock_server_5002):

        retry = Retry(
            total=5,
            backoff_factor=0.5,
            status_forcelist=[429, 500, 502, 503, 504],
        )

        adapter = HTTPAdapter(max_retries=retry)
        s = Session()
        s.mount('http://', adapter)
        url = "http://localhost:5002/graphql"
        req = Request('POST', url, json=mock_payload)
        req_prepped = req.prepare()
        # print("REQUEST:", req_prepped.headers, req_prepped.body)
        resp = s.send(req_prepped)
        # print("RESPONSE:", resp)
        assert resp.status_code == 200
        resp = resp.json()
        assert resp.keys() == {"data", "errors"}
        assert resp["errors"][0]["message"] == "User is not authenticated"
